# Author: Nathan Trouvain at 06/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from reservoirpy.nodes import FORCE, Reservoir


def test_force_init():
    with pytest.warns(DeprecationWarning):
        node = FORCE(10)

    data = np.ones((1, 100))
    res = node(data)

    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (1, 10)
    assert node.alpha == 1e-6

    data = np.ones((10000, 100))
    res = node.run(data)

    assert res.shape == (10000, 10)


def test_force_train_one_step():

    with pytest.warns(DeprecationWarning):
        node = FORCE(10)

    x = np.ones((5, 2))
    y = np.ones((5, 10))

    for x, y in zip(x, y):
        res = node.train(x, y)

    assert node.Wout.shape == (2, 10)
    assert node.bias.shape == (1, 10)
    assert node.alpha == 1e-6

    data = np.ones((10000, 2))
    res = node.run(data)

    assert res.shape == (10000, 10)


def test_force_train():

    with pytest.warns(DeprecationWarning):
        node = FORCE(10)

    X, Y = np.ones((200, 100)), np.ones((200, 10))

    res = node.train(X, Y)

    assert res.shape == (200, 10)
    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.01, decimal=4)
    assert node.bias.shape == (1, 10)
    assert_array_almost_equal(node.bias, np.ones((1, 10)) * 0.01, decimal=4)

    with pytest.warns(DeprecationWarning):
        node = FORCE(10)

    X, Y = np.ones((200, 100)), np.ones((200, 10))

    res = node.train(X, Y)

    assert res.shape == (200, 10)
    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.01, decimal=4)
    assert node.bias.shape == (1, 10)
    assert_array_almost_equal(node.bias, np.ones((1, 10)) * 0.01, decimal=4)

    with pytest.warns(DeprecationWarning):
        node = FORCE(10)

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))

    for x, y in zip(X, Y):
        res = node.train(x, y)

    assert node.Wout.shape == (100, 10)
    assert_array_almost_equal(node.Wout, np.ones((100, 10)) * 0.01, decimal=4)
    assert node.bias.shape == (1, 10)
    assert_array_almost_equal(node.bias, np.ones((1, 10)) * 0.01, decimal=4)

    data = np.ones((10000, 100))
    res = node.run(data)

    assert res.shape == (10000, 10)


def test_force_lms():

    with pytest.warns(DeprecationWarning):
        node = FORCE(10, rule="lms")

    X, Y = np.ones((200, 100)), np.ones((200, 10))

    res = node.train(X, Y)

    assert res.shape == (200, 10)
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (1, 10)

    def alpha_gen():
        while True:
            yield 1e-6

    with pytest.warns(DeprecationWarning):
        node = FORCE(10, rule="lms", alpha=alpha_gen())

    X, Y = np.ones((200, 100)), np.ones((200, 10))

    res = node.train(X, Y)

    assert res.shape == (200, 10)
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (1, 10)

    with pytest.warns(DeprecationWarning):
        node = FORCE(10, rule="lms", alpha=alpha_gen())

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))

    for x, y in zip(X, Y):
        res = node.train(x, y, learn_every=5)

    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (1, 10)

    data = np.ones((10000, 100))
    res = node.run(data)

    assert res.shape == (10000, 10)


def test_esn_force():

    with pytest.warns(DeprecationWarning):
        readout = FORCE(10)
    reservoir = Reservoir(100)

    esn = reservoir >> readout

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))

    for x, y in zip(X, Y):
        res = esn.train(x, y)

    assert readout.Wout.shape == (100, 10)
    assert readout.bias.shape == (1, 10)

    data = np.ones((10000, 100))
    res = esn.run(data)

    assert res.shape == (10000, 10)


def test_esn_force_lms():
    
    with pytest.warns(DeprecationWarning):
        readout = FORCE(10, rule="lms")
    reservoir = Reservoir(100)

    esn = reservoir >> readout

    X, Y = np.ones((5, 200, 100)), np.ones((5, 200, 10))

    for x, y in zip(X, Y):
        res = esn.train(x, y)

    assert readout.Wout.shape == (100, 10)
    assert readout.bias.shape == (1, 10)

    data = np.ones((10000, 100))
    res = esn.run(data)

    assert res.shape == (10000, 10)


def test_force_feedback():

    with pytest.warns(DeprecationWarning):
        readout = FORCE(10)
    reservoir = Reservoir(100)

    esn = reservoir >> readout

    reservoir <<= readout

    X, Y = np.ones((5, 200, 8)), np.ones((5, 200, 10))
    for x, y in zip(X, Y):
        res = esn.train(x, y)

    assert readout.Wout.shape == (100, 10)
    assert readout.bias.shape == (1, 10)
    assert reservoir.Wfb.shape == (100, 10)

    data = np.ones((10000, 8))
    res = esn.run(data)

    assert res.shape == (10000, 10)


def test_hierarchical_esn():

    with pytest.warns(DeprecationWarning):
        readout1 = FORCE(10, name="r1")
    reservoir1 = Reservoir(100)
    with pytest.warns(DeprecationWarning):
        readout2 = FORCE(3, name="r2")
    reservoir2 = Reservoir(100)

    esn = reservoir1 >> readout1 >> reservoir2 >> readout2

    X, Y = np.ones((200, 5)), {"r1": np.ones((200, 10)), "r2": np.ones((200, 3))}
    res = esn.train(X, Y)

    assert readout1.Wout.shape == (100, 10)
    assert readout1.bias.shape == (1, 10)

    assert readout2.Wout.shape == (100, 3)
    assert readout2.bias.shape == (1, 3)

    assert reservoir1.Win.shape == (100, 5)
    assert reservoir2.Win.shape == (100, 10)

    data = np.ones((10000, 5))
    res = esn.run(data)

    assert res.shape == (10000, 3)


def test_dummy_mutual_supervision():
    with pytest.warns(DeprecationWarning):
        readout1 = FORCE(1, name="r1")
    reservoir1 = Reservoir(100)
    with pytest.warns(DeprecationWarning):
        readout2 = FORCE(1, name="r2")
    reservoir2 = Reservoir(100)

    reservoir1 <<= readout1
    reservoir2 <<= readout2

    branch1 = reservoir1 >> readout1
    branch2 = reservoir2 >> readout2

    model = branch1 & branch2

    X = np.ones((200, 5))

    res = model.train(X, Y={"r1": readout2, "r2": readout1}, force_teachers=True)

    assert readout1.Wout.shape == (100, 1)
    assert readout1.bias.shape == (1, 1)

    assert readout2.Wout.shape == (100, 1)
    assert readout2.bias.shape == (1, 1)

    assert reservoir1.Win.shape == (100, 5)
    assert reservoir2.Win.shape == (100, 5)



# Author: Nathan Trouvain at 22/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

import numpy as np
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix

global_dtype = np.float64
global_ctype = "d"

Weights = TypeVar("Weights", np.ndarray, csr_matrix, csc_matrix, coo_matrix)
Shape = TypeVar("Shape", int, Tuple[int, ...])
Data = TypeVar("Data", Iterable[np.ndarray], np.ndarray)
MappedData = TypeVar(
    "MappedData",
    Iterable[np.ndarray],
    np.ndarray,
    Dict[str, Iterable[np.ndarray]],
    Dict[str, np.ndarray],
)


class NodeType(Protocol):
    """Node base Protocol class for type checking and interface inheritance."""

    name: str
    params: Dict[str, Any]
    hypers: Dict[str, Any]
    is_initialized: bool
    input_dim: Shape
    output_dim: Shape
    is_trained_offline: bool
    is_trained_online: bool
    is_trainable: bool
    fitted: bool

    def __call__(self, *args, **kwargs) -> np.ndarray:
        ...

    def __rshift__(self, other: Union["NodeType", Sequence["NodeType"]]) -> "NodeType":
        ...

    def __rrshift__(self, other: Union["NodeType", Sequence["NodeType"]]) -> "NodeType":
        ...

    def __and__(self, other: Union["NodeType", Sequence["NodeType"]]) -> "NodeType":
        ...

    def get_param(self, name: str) -> Any:
        ...

    def initialize(self, x: MappedData = None, y: MappedData = None):
        ...

    def reset(self, to_state: np.ndarray = None) -> "NodeType":
        ...

    def with_state(
        self, state=None, stateful=False, reset=False
    ) -> Iterator["NodeType"]:
        ...

    def with_feedback(
        self, feedback=None, stateful=False, reset=False
    ) -> Iterator["NodeType"]:
        ...


Activation = Callable[[np.ndarray], np.ndarray]
ForwardFn = Callable[[NodeType, Data], np.ndarray]
BackwardFn = Callable[[NodeType, Optional[Data], Optional[Data]], None]
PartialBackFn = Callable[[NodeType, Data, Optional[Data]], None]
ForwardInitFn = Callable[[NodeType, Optional[Data], Optional[Data]], None]
EmptyInitFn = Callable[[NodeType], None]



"""Simple regression models for readout matrix learning.

This module provides simple linear models that can be used
to compute the readout matrix coefficients with simple
linear regression algorithms, like ridge regularized regression
or any linear model from scikit-learn API.

These models are already packed in the :py:class:`compat.ESN`
class, and can instantiated by passing them as arguments to the `ESN`
object.

In most cases, you won't need to call this module directly. Simply
pass the models to the `ESN` object as parameters.
See the :py:class:`compat.ESN` documentation for more information.
"""
from abc import ABCMeta

import numpy as np
from joblib import Parallel, delayed
from scipy import linalg

from ..type import Data, Weights
from ..utils.parallel import clean_tempfile, get_joblib_backend
from ..utils.validation import add_bias, check_vector
from .utils.parallel import as_memmap


def _solve_ridge(XXT, YXT, ridge):
    return linalg.solve(XXT + ridge, YXT.T, assume_a="sym").T


def _prepare_inputs(X, Y, bias=True, allow_reshape=False):
    if bias:
        X = add_bias(X)
    if not isinstance(X, np.ndarray):
        X = np.vstack(X)
    if not isinstance(Y, np.ndarray):
        Y = np.vstack(Y)

    X = check_vector(X, allow_reshape=allow_reshape)
    Y = check_vector(Y, allow_reshape=allow_reshape)

    return X, Y


def _check_tikhnonv_terms(XXT, YXT, x, y):
    if x.shape[0] != y.shape[0]:
        raise ValueError(
            f"Impossible to perform _ridge regression: dimension mismatch "
            f"between target sequence of shape {y.shape} and state sequence "
            f"of shape {x.shape} ({x.shape[0]} != {y.shape[0]})."
        )

    if y.shape[1] != YXT.shape[0]:
        raise ValueError(
            f"Impossible to perform _ridge regression: dimension mismatch "
            f"between target sequence of shape {y.shape} and expected output "
            f"dimension ({YXT.shape[0]}) ({y.shape[1]} != {YXT.shape[0]})"
        )


class _Model(metaclass=ABCMeta):
    """
    Base template for model learning classes.
    """

    Wout: Weights
    _dim_in: int
    _dim_out: int
    _initialized = False

    @property
    def initialized(self):
        """A boolean indicating wether the internal parameters of the
        model are initialized or not."""
        return self._initialized

    @property
    def dim_in(self):
        """Input dimension of the model (i.e. internal
        states dimension)."""
        return self._dim_in

    @property
    def dim_out(self):
        """Output dimension of the model."""
        return self._dim_out

    def fit(self, X: Data = None, Y: Data = None) -> Weights:
        """Fit states X to targets values Y following the model
        learning rule.

        Parameters
        ----------
        X : numpy.ndarray or list of numpy.ndarray
            Internal states of the reservoir.
        Y : numpy.ndarray or list of numpy.ndarray
            Targets values for each states.
        Returns
        -------
            numpy.ndarray
                A readout matrix of shape (targets dimension,
                states dimension + bias (=1)).
        """
        raise NotImplementedError


class _OfflineModel(_Model, metaclass=ABCMeta):
    def partial_fit(self, X: Data, Y: Data):
        """Partially fit the states X to the targets values
        Y. This method can be used to pre-compute some
        steps of the final fitting method.

        Parameters
        ----------
        X : numpy.ndarray or list of numpy.ndarray
            Internal states of the reservoir.
        Y : numpy.ndarray or list of numpy.ndarray
            Targets values for each states.
        """
        raise NotImplementedError


class RidgeRegression(_OfflineModel):
    """Ridge regression model for reservoir output weights
    learning.

    .. math::

        W_{out} = YX^{T} \\cdot (XX^{T} + \\mathrm{_ridge} \\times \\mathrm{Id}_{_dim_in})

    where :math:`W_out` is the readout matrix learnt through this regression,
    :math:`X` are the internal states, :math:`Y` are the targets vectors,
    and :math:`_dim_in` is the internal state dimension (number of units in the
    reservoir).

    By default, ridge coefficient is set to :math:`0`, which is equivalent to a simple
    analytic resolution using pseudo-inverse.

    Partial fit method allows to concurrently pre-compute :math:`XX^{T]`
    and :math:`YX^{T}` when several independent state sequences are
    provided, for performance, as suggested by [1]_.

    .. [1] Lukosevicius, M. (2012). A Practical Guide to Applying Echo State
           Networks. Neural Networks: Tricks of the Trade.
    """

    def __init__(self, ridge=0.0, workers=-1, dtype=np.float64):
        self.workers = workers

        self._dtype = dtype
        self._ridge = ridge
        self._ridgeid = None
        self._XXT = None
        self._YXT = None

    @property
    def ridge(self):
        """Regularization coefficient of the model."""
        return self._ridge

    @ridge.setter
    def ridge(self, value):
        self._ridge = value
        if self._initialized:
            self._reset_ridge_matrix()

    def _reset_ridge_matrix(self):
        self._ridgeid = self._ridge * np.eye(self._dim_in + 1, dtype=self._dtype)

    def initialize(self, dim_in: int = None, dim_out: int = None):
        """
        Initialize the model internal parameters.

        Parameters
        ----------
        dim_in : int
            States dimension.
        dim_out : int
            Targets dimension.
        """
        if dim_in is not None:
            self._dim_in = dim_in
        if dim_out is not None:
            self._dim_out = dim_out

        if getattr(self, "Wout", None) is None:
            self.Wout = np.zeros((self._dim_in + 1, self._dim_out), dtype=self._dtype)
        if getattr(self, "_XXT", None) is None:
            self._XXT = as_memmap(
                np.zeros((self._dim_in + 1, self._dim_in + 1), dtype=self._dtype),
                caller=self,
            )
        if getattr(self, "_YXT", None) is None:
            self._YXT = as_memmap(
                np.zeros((self._dim_out, self._dim_in + 1), dtype=self._dtype),
                caller=self,
            )
        if getattr(self, "_ridgeid", None) is None:
            self._reset_ridge_matrix()

        self._initialized = True

    def clean(self):
        """Clean all internal parameters of the model."""
        del self._XXT
        del self._YXT
        if self._initialized:
            self.initialize()

    def partial_fit(self, X, Y):
        X, Y = _prepare_inputs(X, Y, allow_reshape=True)

        if not self._initialized:
            raise RuntimeError(
                f"RidgeRegression model was never initialized. Call "
                f"'initialize() first."
            )

        _check_tikhnonv_terms(self._XXT, self._YXT, X, Y)

        xxt = X.T.dot(X)
        yxt = Y.T.dot(X)

        # Lock the memory map to avoid increment from
        # different processes at the same time (Numpy doesn't like that).
        self._XXT += xxt
        self._YXT += yxt

    def fit(self, X=None, Y=None):
        if X is not None and Y is not None:
            if isinstance(X, np.ndarray) and isinstance(Y, np.ndarray):
                self.partial_fit(X, Y)

            elif isinstance(X, list) and isinstance(Y, list):
                # if all states and all teachers are given at once,
                # perform partial fit anyway to avoid memory overload.
                workers = min(self.workers, len(X))
                backend = get_joblib_backend(workers)
                with Parallel(n_jobs=workers, backend=backend) as parallel:
                    parallel(delayed(self.partial_fit)(x, y) for x, y in zip(X, Y))

        self.Wout = _solve_ridge(self._XXT, self._YXT, self._ridgeid)

        clean_tempfile(self)
        self.clean()

        return self.Wout



# Author: Nathan Trouvain at 08/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from .intrinsic_plasticity import IPReservoir
from .local_plasticity_reservoir import LocalPlasticityReservoir
from .nvar import NVAR
from .reservoir import Reservoir

__all__ = ["Reservoir", "IPReservoir", "NVAR", "LocalPlasticityReservoir"]



# Author: Nathan Trouvain at 21/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numbers
from typing import Any, Mapping

import numpy as np
from scipy.sparse import issparse

#
# def check_node_io(
#     node, x, expected_dim, allow_timespans=False, allow_list=True, io_type="input"
# ):
#
#     if io_type == "output":
#         msg = (
#             "Impossible to fit node {}: node expected output "
#             "dimension is (1, {}) and teacher vector dimension is {}."
#         )
#     else:
#         msg = (
#             "Impossible to call node {}: node input dimension is (1, {}) "
#             "and input dimension is {}."
#         )
#
#     if isinstance(x, np.ndarray):
#         x = check_vector(
#             x, allow_reshape=True, allow_timespans=allow_timespans, caller=node
#         )
#
#         if node.is_initialized:
#             if x.shape[1] != expected_dim:
#                 raise ValueError(msg.format(node.name, expected_dim, x.shape))
#
#     elif isinstance(x, list):
#         if allow_list:
#             for i in range(len(x)):
#                 x[i] = check_vector(
#                     x[i],
#                     allow_reshape=True,
#                     allow_timespans=allow_timespans,
#                     caller=node,
#                 )
#         else:
#             raise TypeError(
#                 f"Data type not understood. Expected a single "
#                 f"Numpy array but received list in {node}: {x}"
#             )
#     # if X is a teacher Node
#     elif (
#         io_type == "output"
#         and hasattr(x, "is_initialized")
#         and hasattr(x, "output_dim")
#     ):
#         if x.is_initialized and expected_dim is not None:
#             if x.output_dim != expected_dim:
#                 raise ValueError(msg.format(node.name, expected_dim, x.output_dim))
#     # if X is a teacher Node
#     elif (
#         io_type == "input" and hasattr(x, "is_initialized") and hasattr(x, "output_dim")
#     ):
#         raise ValueError(
#             "Impossible to use a Node as input X. Nodes can only "
#             "be used to generate targets Y."
#         )
#     return x

#
# def check_node_state(node, s):
#     s = check_vector(s, allow_timespans=False, caller=node)
#
#     if not node.is_initialized:
#         raise RuntimeError(
#             f"Impossible to set state of node {node.name}: node "
#             f"is not initialized yet."
#         )
#
#     if s.shape[1] != node.output_dim:
#         raise ValueError(
#             f"Impossible to set state of node {node.name}: "
#             f"dimension mismatch between state vector ("
#             f"{s.shape[1]}) "
#             f"and node output dim ({node.output_dim})."
#         )
#     return s


def is_sequence_set(seq: Any) -> bool:
    return isinstance(seq, list) or (isinstance(seq, np.ndarray) and seq.ndim > 2)


def is_array(obj: Any) -> bool:
    return obj is not None and isinstance(obj, np.ndarray) or issparse(obj)


def is_mapping(obj):
    return isinstance(obj, Mapping) or (
        (hasattr(obj, "items") and hasattr(obj, "get"))
        or (
            not (isinstance(obj, list) or isinstance(obj, tuple))
            and hasattr(obj, "__getitem__")
            and not hasattr(obj, "__array__")
        )
    )


def add_bias(X):
    if isinstance(X, np.ndarray):
        X = np.atleast_2d(X)
        return np.hstack([np.ones((X.shape[0], 1)), X])
    elif isinstance(X, list):
        new_X = []
        for x in X:
            x = np.atleast_2d(x)
            new_X.append(np.hstack([np.ones((x.shape[0], 1)), x]))
        return new_X


def check_vector(array, allow_reshape=True, allow_timespans=True, caller=None):
    msg = "."
    if caller is not None:
        if hasattr(caller, "name"):
            msg = f" in {caller.name}."
        else:
            msg = f"in {caller}."

    if not isinstance(array, np.ndarray):
        # maybe a single number, make it an array
        if isinstance(array, numbers.Number):
            array = np.asarray(array)
        else:
            msg = (
                f"Data type '{type(array)}' not understood. All vectors "
                f"should be Numpy arrays" + msg
            )
            raise TypeError(msg)

    if not (np.issubdtype(array.dtype, np.number)):
        msg = f"Impossible to operate on non-numerical data, in array: {array}" + msg
        raise TypeError(msg)

    if allow_reshape:
        array = np.atleast_2d(array)

    if not allow_timespans:
        if array.shape[0] > 1:
            msg = (
                f"Impossible to operate on multiple timesteps. Data should"
                f" have shape (1, n) but is {array.shape}" + msg
            )
            raise ValueError(msg)

    # TODO: choose axis to expand and/or np.atleast_2d

    return array



# Author: Nathan Trouvain at 10/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from reservoirpy.nodes.io import Input, Output

from ..model import Model
from ..node import Node
from ..ops import merge
from .dummy_nodes import *


def test_node_link(plus_node, minus_node):
    clean_registry(Model)

    model1 = plus_node >> minus_node
    model2 = minus_node >> plus_node

    model1._hypers["hyper1"] = "hyper1"
    model1._params["param1"] = "param1"
    assert model1.name == "Model-0"
    model1.name = "Model-1000"
    assert model1.name == "Model-1000"
    assert model1.params["PlusNode-0"]["c"] is None
    assert model1.hypers["PlusNode-0"]["h"] == 1
    assert model1.hyper1 == "hyper1"
    assert model1.param1 == "param1"
    assert model1["PlusNode-0"].input_dim is None

    assert model2.name == "Model-1"
    with pytest.raises(NameError):  # already taken
        model2.name = "Model-1000"
    assert model2.params["PlusNode-0"]["c"] is None
    assert model2.hypers["PlusNode-0"]["h"] == 1
    assert model2["PlusNode-0"].input_dim is None

    assert model1.edges == [(plus_node, minus_node)]
    assert model2.edges == [(minus_node, plus_node)]
    assert set(model1.nodes) == set(model2.nodes)

    with pytest.raises(RuntimeError):
        model1 & model2

    with pytest.raises(NameError):
        _ = model1.get_param("fake_parameter")

    with pytest.raises(RuntimeError):
        plus_node >> minus_node >> plus_node

    with pytest.raises(RuntimeError):
        plus_node >> plus_node


def test_complex_node_link():
    clean_registry(Model)

    A = Node(name="A")
    B = Node(name="B")
    C = Node(name="C")
    D = Node(name="D")
    E = Node(name="E")
    F = Node(name="F")
    In = Input(name="In")
    Out = Output(name="Out")

    path1, path2 = A >> F, B >> E
    path3 = In >> [A, B, C]
    path4 = A >> B >> C >> D >> E >> F >> Out
    model = path1 & path2 & path3 & path4

    assert len(model.nodes) == 12  # 8 user-defined + 4 concat nodes
    assert len(model.edges) == 15  # 11 user-defined + 4 created connections


def test_empty_model_init():
    model = Model()
    assert model.is_empty


def test_model_call(plus_node, minus_node):
    model = plus_node >> minus_node

    data = np.zeros((1, 5))
    res = model(data)

    assert_array_equal(res, data)

    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = branch1 & branch2

    res = model(data)

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, data + 2)
        else:
            assert_array_equal(arr, data - 2)

    res = model(data)

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, data + 4)
        else:
            assert_array_equal(arr, data)

    res = model(data, reset=True)

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, data + 2)
        else:
            assert_array_equal(arr, data - 2)

    res = model(data, stateful=False)

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, data + 4)
        else:
            assert_array_equal(arr, data)

    for node in model.output_nodes:
        if node.name == "PlusNode-0":
            assert_array_equal(node.state(), data + 2)
        else:
            assert_array_equal(node.state(), data - 2)


def test_model_with_state(plus_node, minus_node):
    model = plus_node >> minus_node

    data = np.zeros((1, 5))
    res = model(data)

    assert_array_equal(res, data)

    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = branch1 & branch2

    res = model(data)

    with model.with_state(state={plus_node.name: np.zeros_like(plus_node.state())}):
        assert_array_equal(plus_node.state(), np.zeros_like(plus_node.state()))

    with pytest.raises(TypeError):
        with model.with_state(state=np.zeros_like(plus_node.state())):
            pass


def test_model_run(plus_node, minus_node):
    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = merge(branch1, branch2)

    data = np.zeros((3, 5))
    res = model.run(data)

    expected_plus = np.array([[2] * 5, [4] * 5, [6] * 5])
    expected_minus = np.array([[-2] * 5, [0] * 5, [-2] * 5])

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, expected_plus)
            assert_array_equal(arr[-1][np.newaxis, :], plus_node.state())
        else:
            assert_array_equal(arr, expected_minus)
            assert_array_equal(arr[-1][np.newaxis, :], minus_node.state())

    res = model.run(data, reset=True)

    expected_plus = np.array([[2] * 5, [4] * 5, [6] * 5])
    expected_minus = np.array([[-2] * 5, [0] * 5, [-2] * 5])

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, expected_plus)
            assert_array_equal(arr[-1][np.newaxis, :], plus_node.state())
        else:
            assert_array_equal(arr, expected_minus)
            assert_array_equal(arr[-1][np.newaxis, :], minus_node.state())

    res = model.run(data, stateful=False)

    expected_plus2 = np.array([[8] * 5, [10] * 5, [12] * 5])
    expected_minus2 = np.array([[0] * 5, [-2] * 5, [0] * 5])

    for name, arr in res.items():
        assert name in [out.name for out in model.output_nodes]
        if name == "PlusNode-0":
            assert_array_equal(arr, expected_plus2)
            assert_array_equal(expected_plus[-1][np.newaxis, :], plus_node.state())
        else:
            assert_array_equal(arr, expected_minus2)
            assert_array_equal(expected_minus[-1][np.newaxis, :], minus_node.state())


def test_model_run_on_sequences(plus_node, minus_node):
    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = branch1 & branch2

    data = np.zeros((5, 3, 5))
    res = model.run(data)

    assert set(res.keys()) == {plus_node.name, minus_node.name}
    assert len(res[plus_node.name]) == 5
    assert len(res[minus_node.name]) == 5
    assert res[plus_node.name][0].shape == (3, 5)

    input_node = Input()
    branch1 = input_node >> plus_node
    branch2 = input_node >> minus_node

    model = branch1 & branch2

    data = [np.zeros((3, 5)), np.zeros((8, 5))]
    res = model.run(data)

    assert set(res.keys()) == {plus_node.name, minus_node.name}
    assert len(res[plus_node.name]) == 2
    assert len(res[minus_node.name]) == 2
    assert res[plus_node.name][0].shape == (3, 5)
    assert res[plus_node.name][1].shape == (8, 5)


def test_model_feedback(plus_node, minus_node, feedback_node):
    model = plus_node >> feedback_node >> minus_node
    feedback_node <<= minus_node

    data = np.zeros((1, 5))
    res = model(data)

    assert_array_equal(res, data + 1)
    assert_array_equal(feedback_node.state(), data + 3)

    res = model(data)
    assert_array_equal(res, data + 3)
    assert_array_equal(feedback_node.state(), data + 6)


def test_model_feedback_run(plus_node, minus_node, feedback_node):
    model = plus_node >> feedback_node >> minus_node
    feedback_node <<= minus_node

    data = np.zeros((3, 5))
    res = model.run(data)

    expected = np.array([[1] * 5, [3] * 5, [5] * 5])

    assert_array_equal(res, expected)
    assert_array_equal(feedback_node.state(), data[0][np.newaxis, :] + 10)


def test_model_feedback_forcing_sender(plus_node, minus_node, feedback_node):
    model = plus_node >> feedback_node >> minus_node
    feedback_node <<= minus_node

    data = np.zeros((3, 5))
    res = model.run(data, forced_feedbacks={"MinusNode-0": data + 1}, shift_fb=False)
    expected = np.array([[2] * 5, [2] * 5, [4] * 5])

    assert_array_equal(res, expected)
    assert_array_equal(feedback_node.state(), data[0][np.newaxis, :] + 8)


def test_model_feedback_forcing_receiver(plus_node, minus_node, feedback_node):
    model = plus_node >> feedback_node >> minus_node
    feedback_node <<= minus_node

    data = np.zeros((3, 5))
    res = model.run(data, forced_feedbacks={"FBNode-0": data + 1}, shift_fb=False)
    expected = np.array([[2] * 5, [2] * 5, [4] * 5])

    assert_array_equal(res, expected)
    assert_array_equal(feedback_node.state(), data[0][np.newaxis, :] + 8)


def test_model_feedback_from_previous_node(plus_node, minus_node, feedback_node):
    model = plus_node >> feedback_node >> minus_node
    feedback_node <<= plus_node  # feedback in time, not in space anymore

    data = np.zeros((3, 5))
    res = model.run(data)

    expected = np.array([[1] * 5, [4] * 5, [5] * 5])

    assert_array_equal(res, expected)
    assert_array_equal(feedback_node.state(), data[0][np.newaxis, :] + 11)


def test_model_feedback_from_outsider(plus_node, feedback_node, inverter_node):
    model = plus_node >> feedback_node
    feedback_node <<= plus_node >> inverter_node

    data = np.zeros((1, 5))
    res = model(data)

    assert_array_equal(res, data + 3)
    assert_array_equal(plus_node.state(), data + 2)
    assert_array_equal(inverter_node.state(), data)

    res = model(data)
    assert_array_equal(res, data + 3)
    assert_array_equal(plus_node.state(), data + 4)
    assert_array_equal(inverter_node.state(), data - 2)


def test_model_feedback_from_outsider_complex(
    plus_node, feedback_node, inverter_node, minus_node
):
    model = plus_node >> feedback_node
    fb_model = plus_node >> inverter_node >> minus_node
    feedback_node <<= fb_model

    data = np.zeros((1, 5))
    res = model(data)

    assert_array_equal(res, data + 1)
    assert_array_equal(plus_node.state(), data + 2)
    assert_array_equal(minus_node.state(), data - 2)

    res = model(data)

    assert_array_equal(res, data + 3)
    assert_array_equal(plus_node.state(), data + 4)
    assert_array_equal(minus_node.state(), data - 2)


def test_offline_fit_simple_model(offline_node, offline_node2, plus_node, minus_node):
    model = plus_node >> offline_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.fit(X, Y)

    assert_array_equal(offline_node.b, np.array([6.5]))

    X = np.ones((3, 5, 5)) * 0.5
    Y = np.ones((3, 5, 5))

    model.fit(X, Y)

    assert_array_equal(offline_node.b, np.array([94.5]))

    model.fit(X, Y, reset=True)

    assert_array_equal(offline_node.b, np.array([19.5]))

    res = model.run(X[0], reset=True)

    exp = np.tile(np.array([22.0, 24.5, 27.0, 29.5, 32.0]), 5).reshape(5, 5).T

    assert_array_equal(exp, res)


def test_offline_fit_simple_model_fb(
    basic_offline_node, offline_node2, plus_node, minus_node, feedback_node
):
    model = plus_node >> feedback_node >> basic_offline_node
    feedback_node <<= basic_offline_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.fit(X, Y)

    assert_array_equal(basic_offline_node.b, np.array([9.3]))

    model = plus_node >> feedback_node >> basic_offline_node
    feedback_node <<= basic_offline_node

    X = np.ones((3, 5, 5)) * 0.5
    Y = np.ones((3, 5, 5))

    model.fit(X, Y)

    assert_array_equal(basic_offline_node.b, np.array([11.4]))

    model.fit(X, Y, reset=True)

    assert_array_equal(basic_offline_node.b, np.array([5.15]))

    res = model.run(X[0], reset=True)

    exp = np.tile(np.array([8.65, 19.8, 33.45, 49.6, 68.25]), 5).reshape(5, 5).T

    assert_array_equal(exp, res)

    model = plus_node >> feedback_node >> basic_offline_node
    feedback_node <<= basic_offline_node

    X = np.ones((3, 5, 5)) * 0.5
    Y = np.ones((3, 5, 5))

    model.fit(X, Y, force_teachers=False)


def test_offline_fit_complex(
    basic_offline_node, offline_node2, plus_node, minus_node, feedback_node
):
    model = [plus_node >> basic_offline_node, plus_node] >> minus_node >> offline_node2

    X = np.ones((5, 5, 5)) * 0.5
    Y_1 = np.ones((5, 5, 5))
    Y_2 = np.ones((5, 5, 10))  # after concat

    model.fit(X, Y={"BasicOffline-0": Y_1, "Offline2-0": Y_2})

    res = model.run(X[0])

    assert res.shape == (5, 10)


def test_online_train_simple(online_node, plus_node):
    model = plus_node >> online_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.train(X, Y)

    assert_array_equal(online_node.b, np.array([42.5]))

    model.train(X, Y, reset=True)

    assert_array_equal(online_node.b, np.array([85]))


def test_online_train_fb_forced(online_node, plus_node, feedback_node):
    model = plus_node >> feedback_node >> online_node

    feedback_node <<= online_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.train(X, Y)

    assert_array_equal(online_node.b, np.array([51.5]))

    model.train(X, Y, reset=True)

    assert_array_equal(online_node.b, np.array([103.0]))


def test_online_train_fb_no_forced(online_node, plus_node, feedback_node):
    model = plus_node >> feedback_node >> online_node

    feedback_node <<= online_node

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.train(X, Y, force_teachers=False)

    assert_array_equal(online_node.b, np.array([189.5]))

    model.train(X, Y, reset=True, force_teachers=False)

    assert_array_equal(online_node.b, np.array([3221.5]))


def test_online_train_teacher_nodes(online_node, plus_node, minus_node):
    X = np.ones((5, 5)) * 0.5
    model = plus_node >> online_node

    with pytest.raises(RuntimeError):
        model.train(X, minus_node)  # Impossible to init node nor infer shape

    model = plus_node >> [minus_node, online_node]

    minus_node.set_output_dim(5)

    model.train(X, minus_node)

    assert_array_equal(online_node.b, np.array([54.0]))

    model.train(X, minus_node, reset=True)

    assert_array_equal(online_node.b, np.array([108.0]))


def test_model_return_states():
    off = Offline(name="offline")
    plus = PlusNode(name="plus")
    minus = MinusNode(name="minus")
    inverter = Inverter(name="inv")

    model = plus >> [minus, off >> inverter]

    X = np.ones((5, 5)) * 0.5
    Y = np.ones((5, 5))

    model.fit(X, Y)

    res = model.run(X)

    assert set(res.keys()) == {"minus", "inv"}

    res = model.run(X, return_states="all")

    assert set(res.keys()) == {"minus", "inv", "offline", "plus"}

    res = model.run(X, return_states=["offline"])

    assert set(res.keys()) == {"offline"}


def test_multiinputs():
    import numpy as np

    from reservoirpy.nodes import Input, Reservoir

    source1, source2 = (
        Input(
            name="s1",
            input_dim=5,
        ),
        Input(
            name="s2",
            input_dim=3,
        ),
    )
    res1, res2 = Reservoir(100), Reservoir(100)
    # model = source1 >> [res1, res2] & source2 >> [res1, res2]
    model = [source1, source2] >> res1 & [source1, source2] >> res2
    outputs = model.run({"s1": np.ones((10, 5)), "s2": np.ones((10, 3))})



"""Chaotic timeseries generators.
"""
# Author: Nathan Trouvain at 2020 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import collections
import os
from typing import Union

import numpy as np
from joblib import Memory
from numpy.random import Generator, RandomState
from scipy.fft import fft, ifft
from scipy.integrate import solve_ivp

from .. import _TEMPDIR
from ..utils.random import rand_generator
from ..utils.validation import check_vector
from ._seed import get_seed

memory = Memory(os.path.join(_TEMPDIR, "datasets"), verbose=0)


def _mg_eq(xt, xtau, a=0.2, b=0.1, n=10):
    """
    Mackey-Glass time delay differential equation, at values x(t) and x(t-tau).
    """
    return -b * xt + a * xtau / (1 + xtau**n)


def _mg_rk4(xt, xtau, a, b, n, h=1.0):
    """
    Runge-Kuta method (RK4) for Mackey-Glass timeseries discretization.
    """
    k1 = h * _mg_eq(xt, xtau, a, b, n)
    k2 = h * _mg_eq(xt + 0.5 * k1, xtau, a, b, n)
    k3 = h * _mg_eq(xt + 0.5 * k2, xtau, a, b, n)
    k4 = h * _mg_eq(xt + k3, xtau, a, b, n)

    return xt + k1 / 6 + k2 / 3 + k3 / 3 + k4 / 6


def henon_map(
    n_timesteps: int,
    a: float = 1.4,
    b: float = 0.3,
    x0: Union[list, np.ndarray] = [0.0, 0.0],
    **kwargs,
) -> np.ndarray:
    """Hénon map discrete timeseries [2]_ [3]_.

    .. math::

        x(n+1) &= 1 - ax(n)^2 + y(n)\\\\
        y(n+1) &= bx(n)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    a : float, default to 1.4
        :math:`a` parameter of the system.
    b : float, default to 0.3
        :math:`b` parameter of the system.
    x0 : array-like of shape (2,), default to [0.0, 0.0]
        Initial conditions of the system.

    Returns
    -------
    array of shape (n_timesteps, 2)
        Hénon map discrete timeseries.

    Examples
    --------
    >>> from reservoirpy.datasets import henon_map
    >>> timeseries = henon_map(2_000, a=1.4, b=0.3)
    >>> timeseries.shape
    (2000, 2)

    .. plot::

        from reservoirpy.datasets import henon_map
        plt.figure()
        timeseries = henon_map(2_000, a=1.4, b=0.3)
        plt.scatter(timeseries[:, 0], timeseries[:, 1], s=1)
        plt.xlabel("X"); plt.ylabel("Y")
        plt.show()

    References
    ----------
    .. [2] M. Hénon, ‘A two-dimensional mapping with a strange
           attractor’, Comm. Math. Phys., vol. 50, no. 1, pp. 69–77, 1976.

    .. [3] `Hénon map <https://en.wikipedia.org/wiki/H%C3%A9non_map>`_
           on Wikipedia

    """
    states = np.zeros((n_timesteps, 2))
    states[0] = np.asarray(x0)

    for i in range(1, n_timesteps):
        states[i][0] = 1 - a * states[i - 1][0] ** 2 + states[i - 1][1]
        states[i][1] = b * states[i - 1][0]

    return states


def logistic_map(
    n_timesteps: int, r: float = 3.9, x0: float = 0.5, **kwargs
) -> np.ndarray:
    """Logistic map discrete timeseries [4]_ [5]_.

    .. math::

        x(n+1) = rx(n)(1-x(n))

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    r : float, default to 3.9
        :math:`r` parameter of the system.
    x0 : float, default to 0.5
        Initial condition of the system.

    Returns
    -------
    array of shape (n_timesteps, 1)
        Logistic map discrete timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import logistic_map
    >>> timeseries = logistic_map(n_timesteps=100)
    >>> timeseries.shape
    (100, 1)

    .. plot::

        from reservoirpy.datasets import logistic_map
        plt.figure()
        timeseries = logistic_map(100)
        plt.plot(timeseries[:100, :])
        plt.xlabel("Timesteps")
        plt.show()

    References
    ----------
    .. [4] R. M. May, ‘Simple mathematical models with very
           complicated dynamics’, Nature, vol. 261, no. 5560,
           Art. no. 5560, Jun. 1976, doi: 10.1038/261459a0.

    .. [5] `Logistic map <https://en.wikipedia.org/wiki/Logistic_map>`_
           on Wikipedia
    """
    if r > 0 and 0 < x0 < 1:
        X = np.zeros(n_timesteps)
        X[0] = x0

        for i in range(1, n_timesteps):
            X[i] = r * X[i - 1] * (1 - X[i - 1])

        return X.reshape(-1, 1)
    elif r <= 0:
        raise ValueError("r should be positive.")
    else:
        raise ValueError("Initial condition x0 should be in ]0;1[.")


def lorenz(
    n_timesteps: int,
    rho: float = 28.0,
    sigma: float = 10.0,
    beta: float = 8.0 / 3.0,
    x0: Union[list, np.ndarray] = [1.0, 1.0, 1.0],
    h: float = 0.03,
    **kwargs,
) -> np.ndarray:
    """Lorenz attractor timeseries as defined by Lorenz in 1963 [6]_ [7]_.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= \\sigma (y-x) \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= x(\\rho - z) - y \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= xy - \\beta z

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    rho : float, default to 28.0
        :math:`\\rho` parameter of the system.
    sigma : float, default to 10.0
        :math:`\\sigma` parameter of the system.
    beta : float, default to 8/3
        :math:`\\beta` parameter of the system.
    x0 : array-like of shape (3,), default to [1.0, 1.0, 1.0]
        Initial conditions of the system.
    h : float, default to 0.03
        Time delta between two discrete timesteps.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Examples
    --------

    >>> from reservoirpy.datasets import lorenz
    >>> timeseries = lorenz(10000)
    >>> timeseries.shape
    (10000, 3)

    .. plot::

        from reservoirpy.datasets import lorenz
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = lorenz(10000)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.1)
        plt.tight_layout()
        plt.show()

    Returns
    -------
    array of shape (n_timesteps, 3)
        Lorenz attractor timeseries.

    References
    ----------
    .. [6] E. N. Lorenz, ‘Deterministic Nonperiodic Flow’,
           Journal of the Atmospheric Sciences, vol. 20, no. 2,
           pp. 130–141, Mar. 1963,
           doi: 10.1175/1520-0469(1963)020<0130:DNF>2.0.CO;2.

    .. [7] `Lorenz system <https://en.wikipedia.org/wiki/Lorenz_system>`_
           on Wikipedia.
    """

    def lorenz_diff(t, state):
        x, y, z = state
        return sigma * (y - x), x * (rho - z) - y, x * y - beta * z

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(lorenz_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def mackey_glass(
    n_timesteps: int,
    tau: int = 17,
    a: float = 0.2,
    b: float = 0.1,
    n: int = 10,
    x0: float = 1.2,
    h: float = 1.0,
    seed: Union[None, int, RandomState, Generator] = None,
    history: Union[None, np.ndarray] = None,
    **kwargs,
) -> np.ndarray:
    """Mackey-Glass timeseries [8]_ [9]_, computed from the Mackey-Glass
    delayed differential equation.

    .. math::

        \\frac{x}{t} = \\frac{ax(t-\\tau)}{1+x(t-\\tau)^n} - bx(t)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to compute.
    tau : int, default to 17
        Time delay :math:`\\tau` of Mackey-Glass equation.
        By defaults, equals to 17. Other values can
        change the chaotic behaviour of the timeseries.
    a : float, default to 0.2
        :math:`a` parameter of the equation.
    b : float, default to 0.1
        :math:`b` parameter of the equation.
    n : int, default to 10
        :math:`n` parameter of the equation.
    x0 : float, optional, default to 1.2
        Initial condition of the timeseries.
    h : float, default to 1.0
        Time delta between two discrete timesteps.
    seed : int or :py:class:`numpy.random.Generator`, optional
        Random state seed for reproducibility.
    history: array of shape (T, ), optional
        Past timesteps used to "warmup" the process.
        T must be greater than tau//h. If None, a random history is generated.


    Examples
    --------

    >>> from reservoirpy.datasets import mackey_glass
    >>> timeseries = mackey_glass(1000)
    >>> timeseries.shape
    (1000, 1)

    .. plot::

        from reservoirpy.datasets import mackey_glass
        plt.figure(figsize=(12, 3))
        timeseries = mackey_glass(500)
        plt.plot(timeseries)
        plt.show()

    Returns
    -------
    array of shape (n_timesteps, 1)
        Mackey-Glass timeseries.

    Note
    ----
        As Mackey-Glass is defined by delayed time differential equations,
        the first timesteps of the timeseries can't be initialized at 0
        (otherwise, the first steps of computation involving these
        not-computed-yet-timesteps would yield inconsistent results).
        A random number generator is therefore used to produce random
        initial timesteps based on the value of the initial condition
        passed as parameter. A default seed is hard-coded to ensure
        reproducibility in any case. It can be changed with the
        :py:func:`set_seed` function.

    References
    ----------
    .. [8] M. C. Mackey and L. Glass, ‘Oscillation and chaos in
           physiological
           control systems’, Science, vol. 197, no. 4300, pp. 287–289,
           Jul. 1977,
           doi: 10.1126/science.267326.

    .. [9] `Mackey-Glass equations
            <https://en.wikipedia.org/wiki/Mackey-Glass_equations>`_
            on Wikipedia.

    """

    history_length = int(np.floor(tau / h))

    if history is None:
        # a random state is needed as the method used to discretize
        # the timeseries needs to use randomly generated initial steps
        # based on the initial condition passed as parameter.
        if seed is None:
            seed = get_seed()

        rs = rand_generator(seed)

        # generate random first step based on the value
        # of the initial condition

        history_ = collections.deque(
            x0 * np.ones(history_length) + 0.2 * (rs.random(history_length) - 0.5)
        )
    else:
        if not history_length <= len(history):
            raise ValueError(
                f"The given history has length of {len(history)} < tau/h"
                f" with tau={tau} and h={h}."
            )
        # use the most recent elements of the provided history
        history_ = collections.deque(history[-history_length:])

    xt = x0

    X = np.zeros(n_timesteps)

    for i in range(0, n_timesteps):
        X[i] = xt

        if tau == 0:
            xtau = 0.0
        else:
            xtau = history_.popleft()
            history_.append(xt)

        xth = _mg_rk4(xt, xtau, a=a, b=b, n=n, h=h)

        xt = xth

    return X.reshape(-1, 1)


def multiscroll(
    n_timesteps: int,
    a: float = 40.0,
    b: float = 3.0,
    c: float = 28.0,
    x0: Union[list, np.ndarray] = [-0.1, 0.5, -0.6],
    h: float = 0.01,
    **kwargs,
) -> np.ndarray:
    """Double scroll attractor timeseries [10]_ [11]_,
    a particular case of multiscroll attractor timeseries.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= a(y - x) \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= (c - a)x - xz + cy \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= xy - bz

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    a : float, default to 40.0
        :math:`a` parameter of the system.
    b : float, default to 3.0
        :math:`b` parameter of the system.
    c : float, default to 28.0
        :math:`c` parameter of the system.
    x0 : array-like of shape (3,), default to [-0.1, 0.5, -0.6]
        Initial conditions of the system.
    h : float, default to 0.01
        Time delta between two discrete timesteps.

    Examples
    --------

    >>> from reservoirpy.datasets import multiscroll
    >>> timeseries = multiscroll(1000)
    >>> timeseries.shape
    (1000, 3)

    .. plot::

        from reservoirpy.datasets import multiscroll
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = multiscroll(10000)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.1)
        plt.tight_layout()
        plt.show()

    Returns
    -------
    array of shape (n_timesteps, 3)
        Multiscroll attractor timeseries.

    References
    ----------
    .. [10] G. Chen and T. Ueta, ‘Yet another chaotic attractor’,
           Int. J. Bifurcation Chaos, vol. 09, no. 07, pp. 1465–1466,
           Jul. 1999, doi: 10.1142/S0218127499001024.

    .. [11] `Chen double scroll attractor
           <https://en.wikipedia.org/wiki/Multiscroll_attractor
           #Chen_attractor>`_
           on Wikipedia.

    """

    def multiscroll_diff(t, state):
        x, y, z = state
        dx = a * (y - x)
        dy = (c - a) * x - x * z + c * y
        dz = x * y - b * z
        return dx, dy, dz

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(
        multiscroll_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs
    )

    return sol.y.T


def doublescroll(
    n_timesteps: int,
    r1: float = 1.2,
    r2: float = 3.44,
    r4: float = 0.193,
    ir: float = 2 * 2.25e-5,
    beta: float = 11.6,
    x0: Union[list, np.ndarray] = [0.37926545, 0.058339, -0.08167691],
    h: float = 0.25,
    **kwargs,
) -> np.ndarray:
    """Double scroll attractor timeseries [10]_ [11]_,
    a particular case of multiscroll attractor timeseries.

    .. math::

        \\frac{\\mathrm{d}V_1}{\\mathrm{d}t} &= \\frac{V_1}{R_1} - \\frac{\\Delta V}{R_2} -
        2I_r \\sinh(\\beta\\Delta V) \\\\
        \\frac{\\mathrm{d}V_2}{\\mathrm{d}t} &= \\frac{\\Delta V}{R_2} +2I_r \\sinh(\\beta\\Delta V) - I\\\\
        \\frac{\\mathrm{d}I}{\\mathrm{d}t} &= V_2 - R_4 I

    where :math:`\\Delta V = V_1 - V_2`.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    r1 : float, default to 1.2
        :math:`R_1` parameter of the system.
    r2 : float, default to 3.44
        :math:`R_2` parameter of the system.
    r4 : float, default to 0.193
        :math:`R_4` parameter of the system.
    ir : float, default to 2*2e.25e-5
        :math:`I_r` parameter of the system.
    beta : float, default to 11.6
        :math:`\\beta` parameter of the system.
    x0 : array-like of shape (3,), default to [0.37926545, 0.058339, -0.08167691]
        Initial conditions of the system.
    h : float, default to 0.01
        Time delta between two discrete timesteps.

    Returns
    -------
    array of shape (n_timesteps, 3)
        Multiscroll attractor timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import doublescroll
    >>> timeseries = doublescroll(1000)
    >>> timeseries.shape
    (1000, 3)

    .. plot::

        from reservoirpy.datasets import doublescroll
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = doublescroll(10000)
        print(timeseries.shape)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.1)
        plt.tight_layout()
        plt.show()


    References
    ----------
    .. [10] G. Chen and T. Ueta, ‘Yet another chaotic attractor’,
           Int. J. Bifurcation Chaos, vol. 09, no. 07, pp. 1465–1466,
           Jul. 1999, doi: 10.1142/S0218127499001024.

    .. [11] `Chen double scroll attractor
           <https://en.wikipedia.org/wiki/Multiscroll_attractor
           #Chen_attractor>`_
           on Wikipedia.
    """

    def doublescroll_diff(t, state):
        V1, V2, i = state

        dV = V1 - V2
        factor = (dV / r2) + ir * np.sinh(beta * dV)
        dV1 = (V1 / r1) - factor
        dV2 = factor - i
        dI = V2 - r4 * i

        return dV1, dV2, dI

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(
        doublescroll_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs
    )

    return sol.y.T


def rabinovich_fabrikant(
    n_timesteps: int,
    alpha: float = 1.1,
    gamma: float = 0.89,
    x0: Union[list, np.ndarray] = [-1, 0, 0.5],
    h: float = 0.05,
    **kwargs,
) -> np.ndarray:
    """Rabinovitch-Fabrikant system [12]_ [13]_ timeseries.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= y(z - 1 + x^2) + \\gamma x \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= x(3z + 1 - x^2) + \\gamma y \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= -2z(\\alpha + xy)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    alpha : float, default to 1.1
        :math:`\\alpha` parameter of the system.
    gamma : float, default to 0.89
        :math:`\\gamma` parameter of the system.
    x0 : array-like of shape (3,), default to [-1, 0, 0.5]
        Initial conditions of the system.
    h : float, default to 0.05
        Time delta between two discrete timesteps.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Returns
    -------
    array of shape (n_timesteps, 3)
        Rabinovitch-Fabrikant system timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import rabinovich_fabrikant
    >>> timeseries = rabinovich_fabrikant(2500)
    >>> timeseries.shape
    (2500, 3)

    .. plot::

        from reservoirpy.datasets import rabinovich_fabrikant
        plt.figure()
        plt.subplot(projection="3d")
        timeseries = rabinovich_fabrikant(2500)
        print(timeseries.shape)
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=0.4)
        plt.tight_layout()
        plt.show()

    References
    ----------
    .. [12] M. I. Rabinovich and A. L. Fabrikant,
           ‘Stochastic self-modulation of waves in
           nonequilibrium media’, p. 8, 1979.

    .. [13] `Rabinovich-Fabrikant equations
           <https://en.wikipedia.org/wiki/Rabinovich%E2%80
           %93Fabrikant_equations>`_
           on Wikipedia.

    """

    def rabinovich_fabrikant_diff(t, state):
        x, y, z = state
        dx = y * (z - 1 + x**2) + gamma * x
        dy = x * (3 * z + 1 - x**2) + gamma * y
        dz = -2 * z * (alpha + x * y)
        return dx, dy, dz

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(
        rabinovich_fabrikant_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs
    )

    return sol.y.T


def narma(
    n_timesteps: int,
    order: int = 30,
    a1: float = 0.2,
    a2: float = 0.04,
    b: float = 1.5,
    c: float = 0.001,
    x0: Union[list, np.ndarray] = [0.0],
    seed: Union[int, RandomState] = None,
    u: np.ndarray = None,
) -> np.ndarray:
    """Non-linear Autoregressive Moving Average (NARMA) timeseries,
    as first defined in [14]_, and as used in [15]_.

    NARMA n-th order dynamical system is defined by the recurrent relation:

    .. math::

        y[t+1] = a_1 y[t] + a_2 y[t] (\\sum_{i=0}^{n-1} y[t-i]) + b u[t-(
        n-1)]u[t] + c

    where :math:`u[t]` are sampled following a uniform distribution in
    :math:`[0, 0.5]`.

    Note
    ----
    In most reservoir computing benchmarks, $u$ is given as an input. If you want
    access to its value, you should create the `u` array and pass it as an argument
    to the function as shown below.
    This choice is made to avoid breaking changes. In future ReservoirPy versions, `u`
    will be returned with `y`.
    See `related discussion <https://github.com/reservoirpy/reservoirpy/issues/142>`_.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    order: int, default to 30
        Order of the system.
    a1 : float, default to 0.2
        :math:`a_1` parameter of the system.
    a2 : float, default to 0.04
        :math:`a_2` parameter of the system.
    b : float, default to 1.5
        :math:`b` parameter of the system.
    c : float, default to 0.001
        :math:`c` parameter of the system.
    x0 : array-like of shape (init_steps,), default to [0.0]
        Initial conditions of the system.
    seed : int or :py:class:`numpy.random.Generator`, optional
        Random state seed for reproducibility.
    u : array of shape (`n_timesteps` + `order`, 1), default to None.
        Input timeseries (usually uniformly distributed). See above note.

    Returns
    -------
    array of shape (n_timesteps, 1)
        NARMA timeseries.

    Example
    -------
    >>> import numpy as np
    >>> from reservoirpy.nodes import Reservoir, Ridge
    >>> from reservoirpy.datasets import narma
    >>> model = Reservoir(100) >> Ridge()
    >>> n_timesteps, order = 2000, 30
    >>> rng = np.random.default_rng(seed=2341)
    >>> u = rng.uniform(0, 0.5, size=(n_timesteps + order, 1))
    >>> y = narma(n_timesteps=n_timesteps, order=order, u=u)
    >>> model = model.fit(u[order:], y)

    .. plot::

        from reservoirpy.datasets import narma
        plt.figure(figsize=(12, 3))
        timeseries = narma(100, order=30)
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [14] A. F. Atiya and A. G. Parlos, ‘New results on recurrent
           network training: unifying the algorithms and accelerating
           convergence,‘ in IEEE Transactions on Neural Networks,
           vol. 11, no. 3, pp. 697-709, May 2000,
           doi: 10.1109/72.846741.

    .. [15] B.Schrauwen, M. Wardermann, D. Verstraeten, J. Steil,
           D. Stroobandt, ‘Improving reservoirs using intrinsic
           plasticity‘,
           Neurocomputing, 71. 1159-1171, 2008,
           doi: 10.1016/j.neucom.2007.12.020.
    """
    if seed is None:
        seed = get_seed()
    rs = rand_generator(seed)

    y = np.zeros((n_timesteps + order, 1))

    x0 = check_vector(np.atleast_2d(np.asarray(x0)))
    y[: x0.shape[0], :] = x0

    if u is None:
        u = rs.uniform(0, 0.5, size=(n_timesteps + order, 1))

    for t in range(order, n_timesteps + order - 1):
        y[t + 1] = (
            a1 * y[t]
            + a2 * y[t] * np.sum(y[t - order : t])
            + b * u[t - order] * u[t]
            + c
        )
    return y[order:, :]


def lorenz96(
    n_timesteps: int,
    warmup: int = 0,
    N: int = 36,
    F: float = 8.0,
    dF: float = 0.01,
    h: float = 0.01,
    x0: Union[list, np.ndarray] = None,
    **kwargs,
) -> np.ndarray:
    """Lorenz96 attractor timeseries as defined by Lorenz in 1996 [17]_.

    .. math::

        \\frac{\\mathrm{d}x_i}{\\mathrm{d} t} = (x_{i+1} - x_{i-2}) x_{i-1} - x_i + F

    where :math:`i = 1, \\dots, N` and :math:`x_{-1} = x_{N-1}`
    and :math:`x_{N+1} = x_1` and :math:`N \\geq 4`.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    warmup : int, default to 0
        Number of timesteps to discard at the beginning of the signal, to remove
        transient states.
    N: int, default to 36
        Dimension of the system.
    F : float, default to F
        :math:`F` parameter of the system.
    dF : float, default to 0.01
        Perturbation applied to initial condition if x0 is None.
    h : float, default to 0.01
        Time delta between two discrete timesteps.
    x0 : array-like of shape (N,), default to None
        Initial conditions of the system. If None, the array is initialized to
        an array of shape (N, ) with value F, except for the first value of the
        array that takes the value F + dF.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Returns
    -------
    array of shape (n_timesteps - warmup, N)
        Lorenz96 timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import lorenz96
    >>> timeseries = lorenz96(1000, h=0.1, N=5)
    >>> timeseries.shape
    (1000, 5)

    .. plot::

        from reservoirpy.datasets import lorenz96
        timeseries = lorenz96(1000, h=0.1, N=5)
        plt.figure()
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [17] Lorenz, E. N. (1996, September).
            Predictability: A problem partly solved. In Proc.
            Seminar on predictability (Vol. 1, No. 1).
    """
    if N < 4:
        raise ValueError("N must be >= 4.")

    if x0 is None:
        x0 = F * np.ones(N)
        x0[0] = F + dF

    if len(x0) != N:
        raise ValueError(
            f"x0 should have shape ({N},), but have shape {np.asarray(x0).shape}"
        )

    def lorenz96_diff(t, state):
        ds = np.zeros(N)
        for i in range(N):
            ds[i] = (state[(i + 1) % N] - state[i - 2]) * state[i - 1] - state[i] + F
        return ds

    t_max = (warmup + n_timesteps) * h

    t_eval = np.linspace(0.0, t_max * h, n_timesteps)

    sol = solve_ivp(
        lorenz96_diff,
        y0=x0,
        t_span=(0.0, t_max * h),
        t_eval=t_eval,
        **kwargs,
    )

    return sol.y.T[warmup:]


def rossler(
    n_timesteps: int,
    a: float = 0.2,
    b: float = 0.2,
    c: float = 5.7,
    x0: Union[list, np.ndarray] = [-0.1, 0.0, 0.02],
    h: float = 0.1,
    **kwargs,
) -> np.ndarray:
    """Rössler attractor timeseries [18]_.

    .. math::

        \\frac{\\mathrm{d}x}{\\mathrm{d}t} &= -y - z \\\\
        \\frac{\\mathrm{d}y}{\\mathrm{d}t} &= x + a y \\\\
        \\frac{\\mathrm{d}z}{\\mathrm{d}t} &= b + z (x - c)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    a : float, default to 0.2
        :math:`a` parameter of the system.
    b : float, default to 0.2
        :math:`b` parameter of the system.
    c : float, default to 5.7
        :math:`c` parameter of the system.
    x0 : array-like of shape (3,), default to [-0.1, 0.0, 0.02]
        Initial conditions of the system.
    h : float, default to 0.1
        Time delta between two discrete timesteps.
    **kwargs:
        Other parameters to pass to the `scipy.integrate.solve_ivp`
        solver.

    Returns
    -------
    array of shape (n_timesteps, 3)
        Rössler attractor timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import rossler
    >>> timeseries = rossler(2000)
    >>> timeseries.shape
    (2000, 3)

    .. plot::

        from reservoirpy.datasets import rossler
        timeseries = rossler(2000)
        plt.figure()
        plt.subplot(projection="3d")
        plt.plot(timeseries[:, 0], timeseries[:, 1], timeseries[:, 2], linewidth=1)
        plt.tight_layout()
        plt.show()


    References
    ----------

    .. [18] O.E. Rössler, "An equation for continuous chaos", Physics Letters A,
            vol 57, Issue 5, Pages 397-398, ISSN 0375-9601, 1976,
            https://doi.org/10.1016/0375-9601(76)90101-8.
    """
    if len(x0) != 3:
        raise ValueError(
            f"x0 should have shape (3,), but have shape {np.asarray(x0).shape}"
        )

    def rossler_diff(t, state):
        x, y, z = state
        dx = -y - z
        dy = x + a * y
        dz = b + z * (x - c)
        return dx, dy, dz

    t_max = n_timesteps * h

    t_eval = np.linspace(0.0, t_max, n_timesteps)

    sol = solve_ivp(rossler_diff, y0=x0, t_span=(0.0, t_max), t_eval=t_eval, **kwargs)

    return sol.y.T


def _kuramoto_sivashinsky_etdrk4(v, *, g, E, E2, Q, f1, f2, f3):
    """A single step of EDTRK4 to solve Kuramoto-Sivashinsky equation.

    Kassam, A. K., & Trefethen, L. N. (2005). Fourth-order time-stepping for stiff PDEs.
    SIAM Journal on Scientific Computing, 26(4), 1214-1233.
    """

    Nv = g * fft(np.real(ifft(v)) ** 2)
    a = E2 * v + Q * Nv
    Na = g * fft(np.real(ifft(a)) ** 2)
    b = E2 * v + Q * Na
    Nb = g * fft(np.real(ifft(b)) ** 2)
    c = E2 * a + Q * (2 * Nb - Nv)
    Nc = g * fft(np.real(ifft(c)) ** 2)
    v = E * v + Nv * f1 + 2 * (Na + Nb) * f2 + Nc * f3

    return v


@memory.cache
def _kuramoto_sivashinsky(n_timesteps, *, warmup, N, M, x0, h):
    # initial conditions
    v0 = fft(x0)

    # ETDRK4 scalars
    k = np.conj(np.r_[np.arange(0, N / 2), [0], np.arange(-N / 2 + 1, 0)]) / M

    L = k**2 - k**4

    E = np.exp(h * L)
    E2 = np.exp(h * L / 2)

    r = np.exp(1j * np.pi * (np.arange(1, M + 1) - 0.5) / M)
    LR = h * np.transpose(np.repeat([L], M, axis=0)) + np.repeat([r], N, axis=0)

    Q = h * np.real(np.mean((np.exp(LR / 2) - 1) / LR, axis=1))

    f1 = (-4 - LR + np.exp(LR) * (4 - 3 * LR + LR**2)) / LR**3
    f1 = h * np.real(np.mean(f1, axis=1))

    f2 = (2 + LR + np.exp(LR) * (-2 + LR)) / LR**3
    f2 = h * np.real(np.mean(f2, axis=1))

    f3 = (-4 - 3 * LR - LR**2 + np.exp(LR) * (4 - LR)) / LR**3
    f3 = h * np.real(np.mean(f3, axis=1))

    g = -0.5j * k

    # integration using ETDRK4 method
    v = np.zeros((n_timesteps, N), dtype=complex)
    v[0] = v0
    for n in range(1, n_timesteps):
        v[n] = _kuramoto_sivashinsky_etdrk4(
            v[n - 1], g=g, E=E, E2=E2, Q=Q, f1=f1, f2=f2, f3=f3
        )

    return np.real(ifft(v[warmup:]))


def kuramoto_sivashinsky(
    n_timesteps: int,
    warmup: int = 0,
    N: int = 128,
    M: float = 16,
    x0: Union[list, np.ndarray] = None,
    h: float = 0.25,
) -> np.ndarray:
    """Kuramoto-Sivashinsky oscillators [19]_ [20]_ [21]_.

    .. math::

        y_t = -yy_x - y_{xx} - y_{xxxx}, ~~ x \\in [0, 32\\pi]

    This 1D partial differential equation is solved using ETDRK4
    (Exponential Time-Differencing 4th order Runge-Kutta) method, as described in [22]_.

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate.
    warmup : int, default to 0
        Number of timesteps to discard at the beginning of the signal, to remove
        transient states.
    N : int, default to 128
        Dimension of the system.
    M : float, default to 0.2
        Number of points for complex means. Modify behaviour of the resulting
        multivariate timeseries.
    x0 : array-like of shape (N,), default to None.
        Initial conditions of the system. If None, x0 is equal to
        :math:`\\cos (\\frac{y}{M}) * (1 + \\sin(\\frac{y}{M}))`
        with :math:`y = 2M\\pi x / N, ~~ x \\in [1, N]`.
    h : float, default to 0.25
        Time delta between two discrete timesteps.

    Returns
    -------
    array of shape (n_timesteps - warmup, N)
        Kuramoto-Sivashinsky equation solution.

    Examples
    --------

    >>> from reservoirpy.datasets import kuramoto_sivashinsky
    >>> timeseries = kuramoto_sivashinsky(1000)
    >>> print(timeseries.shape)

    .. plot::

        from reservoirpy.datasets import kuramoto_sivashinsky
        timeseries = kuramoto_sivashinsky(1000)
        plt.figure(figsize=(12, 4))
        plt.imshow(timeseries.T)
        plt.ylabel("Dimension No.")
        plt.xlabel("Time")
        plt.show()

    References
    ----------

    .. [19] Kuramoto, Y. (1978). Diffusion-Induced Chaos in Reaction Systems.
            Progress of Theoretical Physics Supplement, 64, 346–367.
            https://doi.org/10.1143/PTPS.64.346

    .. [20] Sivashinsky, G. I. (1977). Nonlinear analysis of hydrodynamic instability
            in laminar flames—I. Derivation of basic equations.
            Acta Astronautica, 4(11), 1177–1206.
            https://doi.org/10.1016/0094-5765(77)90096-0

    .. [21] Sivashinsky, G. I. (1980). On Flame Propagation Under Conditions
            of Stoichiometry. SIAM Journal on Applied Mathematics, 39(1), 67–82.
            https://doi.org/10.1137/0139007

    .. [22] Kassam, A. K., & Trefethen, L. N. (2005).
            Fourth-order time-stepping for stiff PDEs.
            SIAM Journal on Scientific Computing, 26(4), 1214-1233.
    """
    if x0 is None:
        x = 2 * M * np.pi * np.arange(1, N + 1) / N
        x0 = np.cos(x / M) * (1 + np.sin(x / M))
    else:
        if not np.asarray(x0).shape[0] == N:
            raise ValueError(
                f"Initial condition x0 should be of shape {N} (= N) but "
                f"has shape {np.asarray(x0).shape}"
            )
        else:
            x0 = np.asarray(x0)

    return _kuramoto_sivashinsky(n_timesteps, warmup=warmup, N=N, M=M, x0=x0, h=h)



# Author: Nathan Trouvain at 10/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest

from reservoirpy.node import Node


def plus_forward(node: Node, x: np.ndarray):
    return x + node.c + node.h + node.state()


def plus_initialize(node: Node, x=None, **kwargs):
    node.set_input_dim(x.shape[1])
    node.set_output_dim(x.shape[1])
    node.set_param("c", 1)


class PlusNode(Node):
    def __init__(self, **kwargs):
        super().__init__(
            params={"c": None},
            hypers={"h": 1},
            forward=plus_forward,
            initializer=plus_initialize,
            **kwargs,
        )


def minus_forward(node: Node, x):
    return x - node.c - node.h - node.state()


def minus_initialize(node: Node, x=None, **kwargs):
    node.set_input_dim(x.shape[1])
    node.set_output_dim(x.shape[1])
    node.set_param("c", 1)


class MinusNode(Node):
    def __init__(self, **kwargs):
        super().__init__(
            params={"c": None},
            hypers={"h": 1},
            forward=minus_forward,
            initializer=minus_initialize,
            **kwargs,
        )


def fb_forward(node: Node, x):
    return node.feedback() + x + 1


def fb_initialize(node: Node, x=None, **kwargs):
    node.set_input_dim(x.shape[1])
    node.set_output_dim(x.shape[1])


def fb_initialize_fb(node: Node, fb=None):
    node.set_feedback_dim(fb.shape[1])


class FBNode(Node):
    def __init__(self, **kwargs):
        super().__init__(
            initializer=fb_initialize,
            fb_initializer=fb_initialize_fb,
            forward=fb_forward,
            **kwargs,
        )


def inv_forward(node: Node, x):
    return -x


def inv_initialize(node: Node, x=None, **kwargs):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


class Inverter(Node):
    def __init__(self, **kwargs):
        super(Inverter, self).__init__(
            initializer=inv_initialize, forward=inv_forward, **kwargs
        )


def off_forward(node: Node, x):
    return x + node.b


def off_partial_backward(node: Node, X_batch, Y_batch=None):
    db = np.mean(np.abs(X_batch - Y_batch))
    b = node.get_buffer("b")
    b += db


def off_backward(node: Node, X=None, Y=None):
    b = node.get_buffer("b")
    node.set_param("b", np.array(b).copy())


def off_initialize(node: Node, x=None, y=None):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


def off_initialize_buffers(node: Node):
    node.create_buffer("b", (1,))


class Offline(Node):
    def __init__(self, **kwargs):
        super(Offline, self).__init__(
            params={"b": 0},
            forward=off_forward,
            partial_backward=off_partial_backward,
            backward=off_backward,
            buffers_initializer=off_initialize_buffers,
            initializer=off_initialize,
            **kwargs,
        )


def off_backward_basic(node: Node, X=None, Y=None):
    b = np.mean(node._X)
    node.set_param("b", np.array(b).copy())


class BasicOffline(Node):
    def __init__(self, **kwargs):
        super(BasicOffline, self).__init__(
            params={"b": 0},
            forward=off_forward,
            backward=off_backward_basic,
            initializer=off_initialize,
        )


def off2_forward(node: Node, x):
    return x + node.b


def off2_partial_backward(node: Node, X_batch, Y_batch=None):
    db = np.mean(np.abs(X_batch - Y_batch))
    b = node.get_buffer("b")
    b += db


def off2_backward(node: Node, X=None, Y=None):
    b = node.get_buffer("b")
    node.set_param("b", np.array(b).copy())


def off2_initialize(node: Node, x=None, y=None):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


def off2_initialize_buffers(node: Node):
    node.create_buffer("b", (1,))


class Offline2(Node):
    def __init__(self, **kwargs):
        super(Offline2, self).__init__(
            params={"b": 0},
            forward=off2_forward,
            partial_backward=off2_partial_backward,
            backward=off2_backward,
            initializer=off2_initialize,
            buffers_initializer=off2_initialize_buffers,
            **kwargs,
        )


def sum_forward(node: Node, x):
    if isinstance(x, list):
        x = np.concatenate(x, axis=0)
    return np.sum(x, axis=0)


def sum_initialize(node: Node, x=None, **kwargs):
    if x is not None:
        if isinstance(x, list):
            x = np.concatenate(x, axis=0)
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


class Sum(Node):
    def __init__(self, **kwargs):
        super(Sum, self).__init__(
            forward=sum_forward, initializer=sum_initialize, **kwargs
        )


def unsupervised_forward(node: Node, x):
    return x + node.b


def unsupervised_partial_backward(node: Node, X_batch, Y_batch=None):
    b = np.mean(X_batch)
    node.set_buffer("b", node.get_buffer("b") + b)


def unsupervised_backward(node: Node, X=None, Y=None):
    b = node.get_buffer("b")
    node.set_param("b", np.array(b).copy())


def unsupervised_initialize(node: Node, x=None, y=None):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


def unsupervised_initialize_buffers(node: Node):
    node.create_buffer("b", (1,))


class Unsupervised(Node):
    def __init__(self, **kwargs):
        super(Unsupervised, self).__init__(
            params={"b": 0},
            forward=unsupervised_forward,
            partial_backward=unsupervised_partial_backward,
            backward=unsupervised_backward,
            initializer=unsupervised_initialize,
            buffers_initializer=unsupervised_initialize_buffers,
            **kwargs,
        )


def on_forward(node: Node, x):
    return x + node.b


def on_train(node: Node, x, y=None):
    if y is not None:
        node.set_param("b", node.b + np.mean(x + y))
    else:
        node.set_param("b", node.b + np.mean(x))


def on_initialize(node: Node, x=None, y=None):
    if x is not None:
        node.set_input_dim(x.shape[1])
        node.set_output_dim(x.shape[1])


class OnlineNode(Node):
    def __init__(self, **kwargs):
        super(OnlineNode, self).__init__(
            params={"b": np.array([0])},
            forward=on_forward,
            train=on_train,
            initializer=on_initialize,
            **kwargs,
        )


def multi_forward(node, data):
    return np.concatenate(data, axis=1)


def multi_init(node, x=None, **kwargs):
    if x is not None:
        if isinstance(x, np.ndarray):
            node.set_input_dim(x.shape[1])
            node.set_output_dim(x.shape[1])
        elif hasattr(x, "__iter__"):
            result = multi_forward(node, x)
            node.set_input_dim(tuple([u.shape[1] for u in x]))
            if result.shape[0] > 1:
                node.set_output_dim(result.shape)
            else:
                node.set_output_dim(result.shape[1])


class MultiInput(Node):
    def __init__(self, **kwargs):
        super(MultiInput, self).__init__(
            forward=multi_forward, initializer=multi_init, **kwargs
        )


def clean_registry(node_class):
    node_class._registry = []
    node_class._factory_id = -1


@pytest.fixture(scope="function")
def plus_node():
    clean_registry(PlusNode)
    return PlusNode()


@pytest.fixture(scope="function")
def minus_node():
    clean_registry(MinusNode)
    return MinusNode()


@pytest.fixture(scope="function")
def feedback_node():
    clean_registry(FBNode)
    return FBNode()


@pytest.fixture(scope="function")
def inverter_node():
    clean_registry(Inverter)
    return Inverter()


@pytest.fixture(scope="function")
def offline_node():
    clean_registry(Offline)
    return Offline()


@pytest.fixture(scope="function")
def offline_node2():
    clean_registry(Offline2)
    return Offline2()


@pytest.fixture(scope="function")
def sum_node():
    clean_registry(Sum)
    return Sum()


@pytest.fixture(scope="function")
def unsupervised_node():
    clean_registry(Unsupervised)
    return Unsupervised()


@pytest.fixture(scope="function")
def online_node():
    clean_registry(OnlineNode)
    return OnlineNode()


@pytest.fixture(scope="function")
def basic_offline_node():
    clean_registry(BasicOffline)
    return BasicOffline()


@pytest.fixture(scope="function")
def multiinput():
    clean_registry(MultiInput)
    return MultiInput()



# Author: Nathan Trouvain at 27/09/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np

from ...node import Node
from ...utils.validation import add_bias, check_vector


def _initialize_readout(
    readout, x=None, y=None, init_func=None, bias_init=None, bias=True
):

    if x is not None:

        in_dim = x.shape[1]

        if readout.output_dim is not None:
            out_dim = readout.output_dim
        elif y is not None:
            out_dim = y.shape[1]
        else:
            raise RuntimeError(
                f"Impossible to initialize {readout.name}: "
                f"output dimension was not specified at "
                f"creation, and no teacher vector was given."
            )

        readout.set_input_dim(in_dim)
        readout.set_output_dim(out_dim)

        if callable(init_func):
            W = init_func(in_dim, out_dim, dtype=readout.dtype)
        elif isinstance(init_func, np.ndarray):
            W = (
                check_vector(init_func, caller=readout)
                .reshape(readout.input_dim, readout.output_dim)
                .astype(readout.dtype)
            )
        else:
            raise ValueError(
                f"Data type {type(init_func)} not "
                f"understood for matrix initializer "
                f"'Wout'. It should be an array or "
                f"a callable returning an array."
            )

        if bias:
            if callable(bias_init):
                bias = bias_init(1, out_dim, dtype=readout.dtype)
            elif isinstance(bias_init, np.ndarray):
                bias = (
                    check_vector(bias_init)
                    .reshape(1, readout.output_dim)
                    .astype(readout.dtype)
                )
            else:
                raise ValueError(
                    f"Data type {type(bias_init)} not "
                    f"understood for matrix initializer "
                    f"'bias'. It should be an array or "
                    f"a callable returning an array."
                )
        else:
            bias = np.zeros((1, out_dim), dtype=readout.dtype)

        readout.set_param("Wout", W)
        readout.set_param("bias", bias)


def _prepare_inputs_for_learning(X=None, Y=None, bias=True, allow_reshape=False):
    if X is not None:

        if bias:
            X = add_bias(X)
        if not isinstance(X, np.ndarray):
            X = np.vstack(X)

        X = check_vector(X, allow_reshape=allow_reshape)

    if Y is not None:

        if not isinstance(Y, np.ndarray):
            Y = np.vstack(Y)

        Y = check_vector(Y, allow_reshape=allow_reshape)

    return X, Y


def readout_forward(node: Node, x):
    return (node.Wout.T @ x.reshape(-1, 1) + node.bias.T).T


def _assemble_wout(Wout, bias, has_bias=True):
    wo = Wout
    if has_bias:
        wo = np.r_[bias, wo]
    return wo


def _split_and_save_wout(node, wo):
    if node.input_bias:
        Wout, bias = wo[1:, :], wo[0, :][np.newaxis, :]
        node.set_param("Wout", Wout)
        node.set_param("bias", bias)
    else:
        node.set_param("Wout", wo)


def _compute_error(node, x, y=None):
    """Error between target and prediction."""
    prediction = node.state()
    error = prediction - y
    return error, x.T



# Author: Paul BERNARD at 01/01/2024 <paul.bernard@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from sklearn.decomposition import PCA
from sklearn.gaussian_process.kernels import DotProduct
from sklearn.linear_model import (
    ElasticNet,
    Lars,
    Lasso,
    LassoCV,
    LassoLars,
    LinearRegression,
    LogisticRegression,
    MultiTaskLassoCV,
    OrthogonalMatchingPursuitCV,
    PassiveAggressiveClassifier,
    Perceptron,
    Ridge,
    RidgeClassifier,
    RidgeCV,
    SGDClassifier,
    SGDRegressor,
)
from sklearn.neural_network import MLPClassifier, MLPRegressor

import reservoirpy
from reservoirpy.node import _filter_where_na_target
from reservoirpy.nodes import ScikitLearnNode


def test_fail_non_predictors():
    with pytest.raises(AttributeError):
        _ = ScikitLearnNode(PCA)
    with pytest.raises(AttributeError):
        _ = ScikitLearnNode(DotProduct)


def test_scikitlearn_initializer():

    with pytest.raises(RuntimeError):
        _ = ScikitLearnNode(LinearRegression).initialize()

    with pytest.raises(RuntimeError):
        _ = ScikitLearnNode(LinearRegression).initialize(np.ones((100, 2)))

    _ = ScikitLearnNode(LinearRegression).initialize(
        np.ones((100, 2)), np.ones((100, 2))
    )

    linear_regressor = ScikitLearnNode(
        LinearRegression, output_dim=2, model_hypers={"positive": False}
    ).initialize(np.ones((100, 2)))

    assert linear_regressor.model_hypers == {"positive": False}


# Note that a different seed may fail the tests
@pytest.mark.parametrize(
    "model, model_hypers",
    [
        (LogisticRegression, {"random_state": 2341}),
        (PassiveAggressiveClassifier, {"random_state": 2341}),
        (Perceptron, {"random_state": 2341}),
        (RidgeClassifier, {"random_state": 2341}),
        (SGDClassifier, {"random_state": 2341}),
        (MLPClassifier, {"random_state": 2341}),
    ],
)
def test_scikitlearn_classifiers(model, model_hypers):
    rng = np.random.default_rng(seed=2341)

    X_train = rng.normal(0, 1, size=(10000, 2))
    y_train = (X_train[:, 0:1] > 0.0).astype(np.float16)
    X_test = rng.normal(0, 1, size=(100, 2))
    y_test = (X_test[:, 0:1] > 0.0).astype(np.float16)

    scikit_learn_node = ScikitLearnNode(model=model, model_hypers=model_hypers)

    scikit_learn_node.fit(X_train, y_train)
    y_pred = scikit_learn_node.run(X_test)
    assert y_pred.shape == y_test.shape
    assert np.all(y_pred == y_test)


@pytest.mark.parametrize(
    "model, model_hypers",
    [
        (LinearRegression, None),
        (Ridge, {"random_state": 2341}),
        (SGDRegressor, {"random_state": 2341}),
        (ElasticNet, {"alpha": 1e-4, "random_state": 2341}),
        (Lars, {"random_state": 2341}),
        (Lasso, {"alpha": 1e-4, "random_state": 2341}),
        (LassoLars, {"alpha": 1e-4, "random_state": 2341}),
        (OrthogonalMatchingPursuitCV, {}),
        (MLPRegressor, {"tol": 1e-6, "random_state": 2341}),
    ],
)
def test_scikitlearn_regressors_monooutput(model, model_hypers):
    rng = np.random.default_rng(seed=2341)
    X_train = list(rng.normal(0, 1, size=(30, 100, 2)))
    y_train = [(x[:, 0:1] + x[:, 1:2]).astype(np.float16) for x in X_train]
    X_test = rng.normal(0, 1, size=(100, 2))
    y_test = (X_test[:, 0:1] + X_test[:, 1:2]).astype(np.float16)

    scikit_learn_node = ScikitLearnNode(model=model, model_hypers=model_hypers)

    scikit_learn_node.fit(X_train, y_train)
    y_pred = scikit_learn_node.run(X_test)
    assert y_pred.shape == y_test.shape
    mse = np.mean(np.square(y_pred - y_test))
    assert mse < 2e-4


def test_scikitlearn_multioutput():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(10000, 3))
    y_train = X_train @ np.array([[0, 1, 0], [0, 1, 1], [-1, 0, 1]])
    X_test = rng.normal(0, 1, size=(100, 3))

    lasso = ScikitLearnNode(model=LassoCV, model_hypers={"random_state": 2341}).fit(
        X_train, y_train
    )
    lasso_pred = lasso.run(X_test)

    mt_lasso = ScikitLearnNode(
        model=MultiTaskLassoCV, model_hypers={"random_state": 2341}
    ).fit(X_train, y_train)
    mt_lasso_pred = mt_lasso.run(X_test)

    assert type(lasso.params["instances"]) is list
    assert type(mt_lasso.params["instances"]) is not list

    coef_single = [
        lasso.params["instances"][0].coef_,
        lasso.params["instances"][1].coef_,
        lasso.params["instances"][2].coef_,
    ]
    coef_multitask = mt_lasso.params["instances"].coef_
    assert np.linalg.norm(coef_single[0] - coef_multitask[0]) < 1e-3
    assert np.linalg.norm(coef_single[1] - coef_multitask[1]) < 1e-3
    assert np.linalg.norm(coef_single[2] - coef_multitask[2]) < 1e-3

    assert lasso_pred.shape == mt_lasso_pred.shape == (100, 3)
    assert np.linalg.norm(mt_lasso_pred - lasso_pred) < 1e-2


def test_scikitlearn_reproductibility_random_state():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(100, 3))
    y_train = (X_train @ np.array([0.5, 1, 2])).reshape(-1, 1)
    X_test = rng.normal(0, 1, size=(100, 3))

    # Different scikit-learn random_states
    reservoirpy.set_seed(0)
    y_pred1 = (
        ScikitLearnNode(model=SGDRegressor, model_hypers={"random_state": 1})
        .fit(X_train, y_train)
        .run(X_test)
    )

    reservoirpy.set_seed(0)
    y_pred2 = (
        ScikitLearnNode(model=SGDRegressor, model_hypers={"random_state": 2})
        .fit(X_train, y_train)
        .run(X_test)
    )

    assert not np.all(y_pred1 == y_pred2)

    # Same scikit-learn random_states
    reservoirpy.set_seed(0)
    y_pred1 = (
        ScikitLearnNode(model=SGDRegressor, model_hypers={"random_state": 1})
        .fit(X_train, y_train)
        .run(X_test)
    )

    reservoirpy.set_seed(0)
    y_pred2 = (
        ScikitLearnNode(model=SGDRegressor, model_hypers={"random_state": 1})
        .fit(X_train, y_train)
        .run(X_test)
    )

    assert np.all(y_pred1 == y_pred2)


def test_scikitlearn_reproductibility_rpy_seed():
    rng = np.random.default_rng(seed=2341)
    X_train = rng.normal(0, 1, size=(100, 3))
    y_train = (X_train @ np.array([0.5, 1, 2])).reshape(-1, 1)
    X_test = rng.normal(0, 1, size=(100, 3))

    # Different ReservoirPy random generator
    reservoirpy.set_seed(1)
    y_pred1 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    reservoirpy.set_seed(2)
    y_pred2 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    assert not np.all(y_pred1 == y_pred2)

    # Same ReservoirPy random generator
    reservoirpy.set_seed(0)
    y_pred1 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    reservoirpy.set_seed(0)
    y_pred2 = ScikitLearnNode(model=SGDRegressor).fit(X_train, y_train).run(X_test)

    assert np.all(y_pred1 == y_pred2)


def test_na():
    # Define X, Y, Y_na
    X = np.ones((10, 5))
    Y = np.ones((10, 5))
    Y_na = Y
    Y_na[-1] = np.nan

    readout = ScikitLearnNode(RidgeCV)

    # Filtered the NaN data and get the predictions | Last row removed from X and Y
    X_filtered, Y_filtered = _filter_where_na_target(X, Y_na)
    readout.fit(X_filtered, Y_filtered)
    pred_filtered = readout.run(X_filtered)

    # Manually deleted the last row of the classic inputs
    readout.fit(X[:-1, :], Y[:-1, :])
    pred_classic = readout.run(X[:-1, :])

    # Assert that those predictions are equal
    assert_array_almost_equal(pred_filtered, pred_classic)



# Author: Nathan Trouvain at 08/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import pickle

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from reservoirpy.node import _filter_where_na_target

from ..nodes import Reservoir
from .dummy_nodes import *


def test_node_creation(plus_node):
    assert plus_node.name == "PlusNode-0"
    assert plus_node.params["c"] is None
    assert plus_node.hypers["h"] == 1
    assert plus_node.input_dim is None
    assert plus_node.output_dim is None
    assert not plus_node.is_initialized
    assert hasattr(plus_node, "c")
    assert hasattr(plus_node, "h")
    assert plus_node.state() is None


def test_pickling(plus_node):
    pickled_node = pickle.dumps(plus_node)
    unpickled_node = pickle.loads(pickled_node)

    assert unpickled_node.name == plus_node.name + "-(copy)"
    assert unpickled_node.get_param("h") == plus_node.get_param("h")


def test_node_attr(plus_node):
    assert plus_node.get_param("c") is None
    assert plus_node.get_param("h") == 1

    plus_node.set_param("c", 1)

    assert plus_node.get_param("c") == 1

    with pytest.raises(AttributeError):
        plus_node.get_param("foo")

    with pytest.raises(KeyError):
        plus_node.set_param("foo", 1)

    plus_node.params["a"] = 2
    assert plus_node.get_param("a") == 2

    plus_node.set_param("a", 3)
    assert plus_node.get_param("a") == 3
    assert plus_node.a == 3

    plus_node.a = 4
    assert plus_node.get_param("a") == 4

    assert plus_node.c == 1
    assert plus_node.h == 1

    plus_node.h = 5
    assert plus_node.h == 5


def test_node_init(plus_node):
    data = np.zeros((1, 5))

    res = plus_node(data)

    assert_array_equal(res, data + 2)
    assert plus_node.is_initialized
    assert plus_node.input_dim == 5
    assert plus_node.output_dim == 5
    assert plus_node.c == 1

    data = np.zeros((1, 8))

    with pytest.raises(ValueError):
        plus_node(data)

    with pytest.raises(TypeError):
        plus_node.set_input_dim(9)
    with pytest.raises(TypeError):
        plus_node.set_output_dim(45)


def test_node_init_empty(plus_node):
    plus_noinit = PlusNode(input_dim=5)
    plus_noinit.initialize()

    assert plus_noinit.input_dim == 5
    assert plus_noinit.c == 1
    assert_array_equal(plus_noinit.state(), np.zeros((1, 5)))

    multiinput = MultiInput(input_dim=(5, 2))
    multiinput.initialize()

    assert multiinput.input_dim == (5, 2)

    with pytest.raises(RuntimeError):
        plus_noinit = PlusNode()
        plus_noinit.initialize()

    data = np.zeros((1, 5))

    plus_node.set_input_dim(5)

    plus_node.initialize()

    assert plus_node.is_initialized
    assert plus_node.input_dim == 5
    assert plus_node.output_dim == 5
    assert plus_node.c == 1

    data = np.zeros((1, 8))

    with pytest.raises(ValueError):
        plus_node(data)

    with pytest.raises(TypeError):
        plus_node.set_input_dim(9)
    with pytest.raises(TypeError):
        plus_node.set_output_dim(45)


def test_node_call(plus_node):
    data = np.zeros((1, 5))
    res = plus_node(data)

    assert_array_equal(res, data + 2)
    assert plus_node.state() is not None
    assert_array_equal(data + 2, plus_node.state())

    res2 = plus_node(data)
    assert_array_equal(res2, data + 4)
    assert_array_equal(plus_node.state(), data + 4)

    res3 = plus_node(data, stateful=False)
    assert_array_equal(res3, data + 6)
    assert_array_equal(plus_node.state(), data + 4)

    res4 = plus_node(data, reset=True)
    assert_array_equal(res4, res)
    assert_array_equal(plus_node.state(), data + 2)


def test_node_dimensions(plus_node):
    data = np.zeros((1, 5))
    res = plus_node(data)

    # input size mismatch
    with pytest.raises(ValueError):
        data = np.zeros((1, 6))
        plus_node(data)

    # input size mismatch in run,
    # no matter how many timesteps are given
    with pytest.raises(ValueError):
        data = np.zeros((5, 6))
        plus_node.run(data)

    with pytest.raises(ValueError):
        data = np.zeros((1, 6))
        plus_node.run(data)

    # no timespans in call, only single timesteps
    with pytest.raises(ValueError):
        data = np.zeros((2, 5))
        plus_node(data)


def test_node_state(plus_node):
    data = np.zeros((1, 5))

    with pytest.raises(RuntimeError):
        with plus_node.with_state(np.ones((1, 5))):
            plus_node(data)

    plus_node(data)
    assert_array_equal(plus_node.state(), data + 2)

    with plus_node.with_state(np.ones((1, 5))):
        res_w = plus_node(data)
        assert_array_equal(res_w, data + 3)
    assert_array_equal(plus_node.state(), data + 2)

    with plus_node.with_state(np.ones((1, 5)), stateful=True):
        res_w = plus_node(data)
        assert_array_equal(res_w, data + 3)
    assert_array_equal(plus_node.state(), data + 3)

    with plus_node.with_state(reset=True):
        res_w = plus_node(data)
        assert_array_equal(res_w, data + 2)
    assert_array_equal(plus_node.state(), data + 3)

    with pytest.raises(ValueError):
        with plus_node.with_state(np.ones((1, 8))):
            plus_node(data)


def test_node_run(plus_node):
    data = np.zeros((3, 5))
    res = plus_node.run(data)
    expected = np.array([[2] * 5, [4] * 5, [6] * 5])

    assert_array_equal(res, expected)
    assert_array_equal(res[-1][np.newaxis, :], plus_node.state())

    res2 = plus_node.run(data, stateful=False)
    expected2 = np.array([[8] * 5, [10] * 5, [12] * 5])

    assert_array_equal(res2, expected2)
    assert_array_equal(res[-1][np.newaxis, :], plus_node.state())

    res3 = plus_node.run(data, reset=True)

    assert_array_equal(res3, expected)
    assert_array_equal(res[-1][np.newaxis, :], plus_node.state())


def test_offline_fit(offline_node):
    X = np.ones((10, 5)) * 0.5
    Y = np.ones((10, 5))

    assert offline_node.b == 0

    offline_node.partial_fit(X, Y)

    assert_array_equal(offline_node.get_buffer("b"), np.array([0.5]))

    offline_node.fit()

    assert_array_equal(offline_node.b, np.array([0.5]))

    X = np.ones((10, 5)) * 2.0
    Y = np.ones((10, 5))

    offline_node.fit(X, Y)

    assert_array_equal(offline_node.b, np.array([1.0]))

    X = [np.ones((10, 5)) * 2.0] * 3
    Y = [np.ones((10, 5))] * 3

    offline_node.fit(X, Y)

    assert_array_equal(offline_node.b, np.array([3.0]))

    offline_node.partial_fit(X, Y)

    assert_array_equal(offline_node.get_buffer("b"), np.array([3.0]))


def test_unsupervised_fit(unsupervised_node):
    X = np.ones((10, 5))

    assert unsupervised_node.b == 0

    unsupervised_node.partial_fit(X)

    assert_array_equal(unsupervised_node.get_buffer("b"), np.array([1.0]))

    unsupervised_node.fit()

    assert_array_equal(unsupervised_node.b, np.array([1.0]))

    X = np.ones((10, 5)) * 2.0

    unsupervised_node.fit(X)

    assert_array_equal(unsupervised_node.b, np.array([2.0]))

    X = [np.ones((10, 5)) * 2.0] * 3

    unsupervised_node.fit(X)

    assert_array_equal(unsupervised_node.b, np.array([6.0]))

    unsupervised_node.partial_fit(X)

    assert_array_equal(unsupervised_node.get_buffer("b"), np.array([6.0]))


def test_train_unsupervised(online_node):
    X = np.ones((10, 5))

    assert online_node.b == 0

    online_node.train(X)

    assert_array_equal(online_node.b, np.array([10.0]))

    X = np.ones((10, 5)) * 2.0

    online_node.train(X)

    assert_array_equal(online_node.b, np.array([30.0]))

    X = [np.ones((10, 5)) * 2.0] * 3

    with pytest.raises(TypeError):
        online_node.train(X)


def test_train(online_node):
    X = np.ones((10, 5))
    Y = np.ones((10, 5))

    assert online_node.b == 0

    online_node.train(X, Y)

    assert_array_equal(online_node.b, np.array([20.0]))

    X = np.ones((10, 5)) * 2.0

    online_node.train(X, Y)

    assert_array_equal(online_node.b, np.array([50.0]))

    X = [np.ones((10, 5)) * 2.0] * 3

    with pytest.raises(TypeError):
        online_node.train(X, Y)


def test_train_raise(online_node):
    X = [np.ones((10, 5)) * 2.0] * 3
    Y = [np.ones((10, 5)) * 2.0] * 3

    with pytest.raises(TypeError):
        online_node.train(X, Y)


def test_train_learn_every(online_node):
    X = np.ones((10, 5))
    Y = np.ones((10, 5))

    assert online_node.b == 0

    online_node.train(X, Y, learn_every=2)

    assert_array_equal(online_node.b, np.array([10.0]))

    X = np.ones((10, 5)) * 2.0

    online_node.train(X, Y, learn_every=2)

    assert_array_equal(online_node.b, np.array([25.0]))


def test_train_supervised_by_teacher_node(online_node, plus_node):
    X = np.ones((1, 5))

    # using not initialized node
    with pytest.raises(RuntimeError):
        online_node.train(X, plus_node)

    plus_node(np.ones((1, 5)))

    online_node.train(X, plus_node)

    assert_array_equal(online_node.b, np.array([4.0]))


def test_node_bad_learning_method(online_node, plus_node, offline_node):
    X = np.ones((10, 5))
    Y = np.ones((10, 5))

    with pytest.raises(TypeError):
        online_node.fit(X, Y)

    with pytest.raises(TypeError):
        plus_node.fit(X, Y)

    with pytest.raises(TypeError):
        online_node.partial_fit(X, Y)

    with pytest.raises(TypeError):
        offline_node.train(X, Y)

    with pytest.raises(TypeError):
        plus_node.train(X, Y)


def test_offline_node_bad_warmup(offline_node):
    X = np.ones((10, 5))
    Y = np.ones((10, 5))

    with pytest.raises(ValueError):
        offline_node.fit(X, Y, warmup=10)


def test_offline_node_default_partial(basic_offline_node):
    X = np.ones((10, 5))
    Y = np.ones((10, 5))

    basic_offline_node.partial_fit(X, Y, warmup=2)
    assert_array_equal(basic_offline_node._X[0], X[2:])


def test_multi_input(multiinput):
    multi_noinit = MultiInput(input_dim=(5, 2))
    multi_noinit.initialize()

    with pytest.raises(RuntimeError):
        multi_noinit = MultiInput()
        multi_noinit.initialize()

    x = [np.ones((1, 5)), np.ones((1, 2))]

    r = multiinput(x)

    assert r.shape == (1, 7)
    assert multiinput.input_dim == (5, 2)

    multi_noinit = MultiInput()
    x = [np.ones((2, 5)), np.ones((2, 2))]
    r = multi_noinit.run(x)

    assert multi_noinit.input_dim == (5, 2)


def test_feedback_noinit(feedback_node):
    with pytest.raises(RuntimeError):
        feedback_node.feedback()

    inv_notinit = Inverter(input_dim=5, output_dim=5)

    feedback_node <<= inv_notinit

    data = np.ones((1, 5))

    feedback_node(data)

    assert_array_equal(feedback_node.feedback(), inv_notinit.state())


def test_feedback_initialize_feedback(feedback_node):
    inv_notinit = Inverter(input_dim=5, output_dim=5)
    feedback_node <<= inv_notinit

    data = np.ones((1, 5))

    feedback_node.initialize_feedback()
    res = feedback_node(data)

    fb = feedback_node.feedback()
    inv_state = inv_notinit.state()

    assert_array_equal(inv_state, fb)

    inv_notinit = Inverter(input_dim=5, output_dim=5)
    plus_noinit = PlusNode(input_dim=5, output_dim=5)

    # default feedback initializer (plus_node is not supposed to handle feedback)
    plus_noinit <<= inv_notinit
    plus_noinit.initialize_feedback()
    res = plus_noinit(data)

    fb = plus_noinit.feedback()
    inv_state = inv_notinit.state()

    assert_array_equal(inv_state, fb)


def test_feedback_init_distant_model(feedback_node, plus_node, inverter_node):
    feedback_node <<= plus_node >> inverter_node

    with pytest.raises(RuntimeError):
        feedback_node.initialize_feedback()

    with pytest.raises(RuntimeError):
        data = np.ones((1, 5))
        feedback_node(data)

    plus_node.initialize(data)

    feedback_node.initialize_feedback()

    fb = feedback_node.feedback()
    inv_state = inverter_node.state()

    assert_array_equal(inv_state, fb)


def test_feedback_init_deep_distant_model(
    feedback_node, plus_node, minus_node, inverter_node
):
    feedback_node <<= plus_node >> minus_node >> inverter_node

    with pytest.raises(RuntimeError):
        feedback_node.initialize_feedback()

    with pytest.raises(RuntimeError):
        data = np.ones((1, 5))
        feedback_node(data)

    plus_node.initialize(data)
    feedback_node.initialize_feedback()

    fb = feedback_node.feedback()
    inv_state = inverter_node.state()

    assert_array_equal(inv_state, fb)


def test_filter_na_target():

    # Generate fake x and y
    x = np.ones(10).reshape(5, 2)
    y = np.ones(5).reshape(5, 1)
    # Replace last y by nan
    y[-1, -1] = np.nan

    # Filter the data
    x, y = _filter_where_na_target(x, y)

    # Assert axis 0 of x and y match
    assert x.shape[0] == y.shape[0]

    # Assert no occurences of NaN values in target y
    assert not np.isnan(y).any()


def test_fit_no_nan(plus_node, offline_node):

    # Define X, Y, Y_na
    X = np.ones((10, 5))
    Y = np.ones((10, 5))
    Y_na = Y
    Y_na[-1] = np.nan

    # Get predictions with target NaN (last row)
    esn_na = plus_node >> offline_node
    esn_na.fit(X, Y_na)
    pred_na = esn_na.run(X)

    # Get predictions with classic X and Y
    esn_classic = plus_node >> offline_node
    esn_classic.fit(X, Y)
    pred_classic = esn_classic.run(X)

    # Assert that both predictions are equal | same inputs should yield same outputs
    assert_array_equal(pred_na, pred_classic)

    # Filtered the NaN data and get the predictions | Last row removed from X and Y
    X_filtered, Y_filtered = _filter_where_na_target(X, Y_na)
    esn_filtered = plus_node >> offline_node
    esn_filtered.fit(X_filtered, Y_filtered)
    pred_filtered = esn_filtered.run(X_filtered)

    # Manually deleted the last row of the classic inputs
    esn_classic = plus_node >> offline_node
    esn_classic.fit(X[:-1, :], Y[:-1, :])
    pred_classic = esn_classic.run(X[:-1, :])

    # Assert that those predictions are equal
    assert_array_equal(pred_filtered, pred_classic)

    # Test for correct ValueError raised
    X = np.ones((10, 5, 3))
    Y = np.ones((10, 5, 3))
    with pytest.raises(ValueError):
        X, Y = _filter_where_na_target(X, Y)



# Author: Nathan Trouvain at 16/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import sys

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from functools import partial
from typing import Callable, Dict, Optional, Sequence, Union

import numpy as np

from ..._base import check_xy
from ...activationsfunc import get_function, identity
from ...mat_gen import bernoulli, uniform
from ...node import Unsupervised, _init_with_sequences
from ...type import Weights
from ...utils.random import noise, rand_generator
from ...utils.validation import is_array
from .base import forward_external
from .base import initialize as initialize_base
from .base import initialize_feedback


def gaussian_gradients(x, y, a, mu, sigma, eta):
    """KL loss gradients of neurons with tanh activation (~ Normal(mu, sigma))."""
    sig2 = sigma**2
    delta_b = -eta * (-(mu / sig2) + (y / sig2) * (2 * sig2 + 1 - y**2 + mu * y))
    delta_a = (eta / a) + delta_b * x
    return delta_a, delta_b


def exp_gradients(x, y, a, mu, eta):
    """KL loss gradients of neurons with sigmoid activation
    (~ Exponential(lambda=1/mu))."""
    delta_b = eta * (1 - (2 + (1 / mu)) * y + (y**2) / mu)
    delta_a = (eta / a) + delta_b * x
    return delta_a, delta_b


def apply_gradients(a, b, delta_a, delta_b):
    """Apply gradients on a and b parameters of intrinsic plasticity."""
    a2 = a + delta_a
    b2 = b + delta_b
    return a2, b2


def ip(reservoir, pre_state, post_state):
    """Perform one step of intrinsic plasticity.

    Optimize a and b such that
    post_state = f(a*pre_state+b) ~ Dist(params) where Dist can be normal or
    exponential."""
    a = reservoir.a
    b = reservoir.b
    mu = reservoir.mu
    eta = reservoir.learning_rate

    if reservoir.activation_type == "tanh":
        sigma = reservoir.sigma
        delta_a, delta_b = gaussian_gradients(
            x=pre_state.T, y=post_state.T, a=a, mu=mu, sigma=sigma, eta=eta
        )
    else:  # sigmoid
        delta_a, delta_b = exp_gradients(
            x=pre_state.T, y=post_state.T, a=a, mu=mu, eta=eta
        )

    return apply_gradients(a=a, b=b, delta_a=delta_a, delta_b=delta_b)


def ip_activation(state, *, reservoir, f):
    """Activation of neurons f(a*x+b) where a and b are intrinsic plasticity
    parameters."""
    a, b = reservoir.a, reservoir.b
    return f(a * state + b)


def backward(reservoir: "IPReservoir", X=None, *args, **kwargs):
    for e in range(reservoir.epochs):
        for seq in X:
            for u in seq:
                post_state = reservoir.call(u.reshape(1, -1))
                pre_state = reservoir.internal_state

                a, b = ip(reservoir, pre_state, post_state)

                reservoir.set_param("a", a)
                reservoir.set_param("b", b)


def initialize(reservoir, *args, **kwargs):

    initialize_base(reservoir, *args, **kwargs)

    a = np.ones((reservoir.output_dim, 1))
    b = np.zeros((reservoir.output_dim, 1))

    reservoir.set_param("a", a)
    reservoir.set_param("b", b)


class IPReservoir(Unsupervised):
    """Pool of neurons with random recurrent connexions, tuned using Intrinsic
    Plasticity.

    Intrinsic Plasticity is applied as described in [1]_ and [2]_.

    Reservoir neurons states, gathered in a vector :math:`\\mathbf{x}`, follow
    the update rule below:

    .. math::


        \\mathbf{r}[t+1] = (1 - \\mathrm{lr}) * \\mathbf{r}[t] + \\mathrm{lr}
        * (\\mathbf{W}_{in} \\cdot (\\mathbf{u}[t+1]+c_{in}*\\xi)
         + \\mathbf{W} \\cdot \\mathbf{x}[t]
        + \\mathbf{W}_{fb} \\cdot (g(\\mathbf{y}[t])+c_{fb}*\\xi) + \\mathbf{b}_{in})

    .. math::

        \\mathbf{x}[t+1] = f(\\mathbf{a}*\\mathbf{r}[t+1]+\\mathbf{b}) + c * \\xi

    Parameters :math:`\\mathbf{a}` and :math:`\\mathbf{b}` are updated following two
    different rules:

    - **1.** Neuron activation is tanh:

    In that case, output distribution should be a Gaussian distribution of parameters
    (:math:`\\mu`, :math:`\\sigma`). The learning rule to obtain this output
    distribution is described in [2]_.

    - **2.** Neuron activation is sigmoid:

    In that case, output distribution should be an exponential distribution of
    parameter :math:`\\mu = \\frac{1}{\\lambda}`.
    The learning rule to obtain this output distribution is described in [1]_ and [2]_.

    where:
        - :math:`\\mathbf{x}` is the output activation vector of the reservoir;
        - :math:`\\mathbf{r}` is the internal activation vector of the reservoir;
        - :math:`\\mathbf{u}` is the input timeseries;
        - :math:`\\mathbf{y}` is a feedback vector;
        - :math:`\\xi` is a random noise;
        - :math:`f` and :math:`g` are activation functions.

    :py:attr:`IPReservoir.params` **list:**

    ================== =================================================================
    ``W``              Recurrent weights matrix (:math:`\\mathbf{W}`).
    ``Win``            Input weights matrix (:math:`\\mathbf{W}_{in}`).
    ``Wfb``            Feedback weights matrix (:math:`\\mathbf{W}_{fb}`).
    ``bias``           Input bias vector (:math:`\\mathbf{b}_{in}`).
    ``internal_state``  Internal state (:math:`\\mathbf{r}`).
    ``a``              Gain of reservoir activation (:math:`\\mathbf{a}`).
    ``b``              Bias of reservoir activation (:math:`\\mathbf{b}`).
    ================== =================================================================

    :py:attr:`IPReservoir.hypers` **list:**

    ======================= ========================================================
    ``lr``                  Leaking rate (1.0 by default) (:math:`\\mathrm{lr}`).
    ``sr``                  Spectral radius of ``W`` (optional).
    ``mu``                  Mean of the target distribution (0.0 by default) (:math:`\\mu`).
    ``sigma``               Variance of the target distribution (1.0 by default) (:math:`\\sigma`).
    ``learning_rate``       Learning rate (5e-4 by default).
    ``epochs``              Number of epochs for training (1 by default).
    ``input_scaling``       Input scaling (float or array) (1.0 by default).
    ``fb_scaling``          Feedback scaling (float or array) (1.0 by default).
    ``rc_connectivity``     Connectivity (or density) of ``W`` (0.1 by default).
    ``input_connectivity``  Connectivity (or density) of ``Win`` (0.1 by default).
    ``fb_connectivity``     Connectivity (or density) of ``Wfb`` (0.1 by default).
    ``noise_in``            Input noise gain (0 by default) (:math:`c_{in} * \\xi`).
    ``noise_rc``            Reservoir state noise gain (0 by default) (:math:`c*\\xi`).
    ``noise_fb``            Feedback noise gain (0 by default) (:math:`c_{fb}*\\xi`).
    ``noise_type``          Distribution of noise (normal by default) (:math:`\\xi\\sim\\mathrm{Noise~type}`).
    ``activation``          Activation of the reservoir units (tanh by default) (:math:`f`).
    ``fb_activation``       Activation of the feedback units (identity by default) (:math:`g`).
    ``units``               Number of neuronal units in the reservoir.
    ``noise_generator``     A random state generator.
    ======================= ========================================================

    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    lr : float or array-like of shape (units,), default to 1.0
        Neurons leak rate. Must be in :math:`[0, 1]`.
    sr : float, optional
        Spectral radius of recurrent weight matrix.
    mu : float, default to 0.0
        Mean of the target distribution.
    sigma : float, default to 1.0
        Variance of the target distribution.
    learning_rate : float, default to 5e-4
        Learning rate.
    epochs : int, default to 1
        Number of training iterations.
    input_bias : bool, default to True
        If False, no bias is added to inputs.
    noise_rc : float, default to 0.0
        Gain of noise applied to reservoir activations.
    noise_in : float, default to 0.0
        Gain of noise applied to input inputs.
    noise_fb : float, default to 0.0
        Gain of noise applied to feedback signal.
    noise_type : str, default to "normal"
        Distribution of noise. Must be a Numpy random variable generator
        distribution (see :py:class:`numpy.random.Generator`).
    noise_kwargs : dict, optional
        Keyword arguments to pass to the noise generator, such as `low` and `high`
        values of uniform distribution.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    bias_scaling: float, default to 1.0
        Bias gain.
    fb_scaling : float or array-like of shape (features,), default to 1.0
        Feedback gain. An array of the same dimension as the feedback can be used to
        set up different feedback scaling for each feature.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    rc_connectivity : float, default to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    fb_connectivity : float, default to 0.1
        Connectivity of feedback neurons, i.e. ratio of feedback neurons
        connected to reservoir neurons. Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Input weights matrix or initializer. If a callable (like a function) is
        used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.uniform`
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    Wfb : callable or array-like of shape (units, feedback), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Feedback weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    fb_activation : str or callable, default to :py:func:`~reservoirpy.activationsfunc.identity`
        Feedback activation function.
        - If a str, should be a :py:mod:`~reservoirpy.activationsfunc`
        function name.
        - If a callable, should be an element-wise operator on arrays.
    activation : {"tanh", "sigmoid"}, default to "tanh"
        Reservoir units activation function.
    feedback_dim : int, optional
        Feedback dimension. Can be inferred at first call.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    name : str, optional
        Node name.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.

    References
    ----------
    .. [1] Triesch, J. (2005). A Gradient Rule for the Plasticity of a
           Neuron’s Intrinsic Excitability. In W. Duch, J. Kacprzyk,
           E. Oja, & S. Zadrożny (Eds.), Artificial Neural Networks:
           Biological Inspirations – ICANN 2005 (pp. 65–70).
           Springer. https://doi.org/10.1007/11550822_11

    .. [2] Schrauwen, B., Wardermann, M., Verstraeten, D., Steil, J. J.,
           & Stroobandt, D. (2008). Improving reservoirs using intrinsic
           plasticity. Neurocomputing, 71(7), 1159–1171.
           https://doi.org/10.1016/j.neucom.2007.12.020

    Example
    -------
    >>> from reservoirpy.nodes import IPReservoir
    >>> reservoir = IPReservoir(
    ...                 100, mu=0.0, sigma=0.1, sr=0.95, activation="tanh", epochs=10)

    We can fit the intrinsic plasticity parameters to reach a normal distribution
    of the reservoir activations.
    Using the :py:func:`~reservoirpy.datasets.narma` timeseries:

    >>> from reservoirpy.datasets import narma
    >>> x = narma(1000)
    >>> _ = reservoir.fit(x, warmup=100)
    >>> states = reservoir.run(x)

    .. plot:: ./api/intrinsic_plasticity_example.py

    """

    def __init__(
        self,
        units: int = None,
        sr: Optional[float] = None,
        lr: float = 1.0,
        mu: float = 0.0,
        sigma: float = 1.0,
        learning_rate: float = 5e-4,
        epochs: int = 1,
        input_bias: bool = True,
        noise_rc: float = 0.0,
        noise_in: float = 0.0,
        noise_fb: float = 0.0,
        noise_type: str = "normal",
        noise_kwargs: Dict = None,
        input_scaling: Union[float, Sequence] = 1.0,
        bias_scaling: float = 1.0,
        fb_scaling: Union[float, Sequence] = 1.0,
        input_connectivity: Optional[float] = 0.1,
        rc_connectivity: Optional[float] = 0.1,
        fb_connectivity: Optional[float] = 0.1,
        Win: Union[Weights, Callable] = bernoulli,
        W: Union[Weights, Callable] = uniform,
        Wfb: Union[Weights, Callable] = bernoulli,
        bias: Union[Weights, Callable] = bernoulli,
        feedback_dim: int = None,
        fb_activation: Union[str, Callable] = identity,
        activation: Literal["tanh", "sigmoid"] = "tanh",
        name=None,
        seed=None,
        **kwargs,
    ):
        if units is None and not is_array(W):
            raise ValueError(
                "'units' parameter must not be None if 'W' parameter is not "
                "a matrix."
            )

        if activation not in ["tanh", "sigmoid"]:
            raise ValueError(
                f"Activation '{activation}' must be 'tanh' or 'sigmoid' when "
                "applying intrinsic plasticity."
            )

        rng = rand_generator(seed=seed)
        noise_kwargs = dict() if noise_kwargs is None else noise_kwargs

        super(IPReservoir, self).__init__(
            fb_initializer=partial(
                initialize_feedback,
                Wfb_init=Wfb,
                fb_scaling=fb_scaling,
                fb_connectivity=fb_connectivity,
                seed=seed,
            ),
            params={
                "W": None,
                "Win": None,
                "Wfb": None,
                "bias": None,
                "a": None,
                "b": None,
                "internal_state": None,
            },
            hypers={
                "sr": sr,
                "lr": lr,
                "mu": mu,
                "sigma": sigma,
                "learning_rate": learning_rate,
                "epochs": epochs,
                "input_bias": input_bias,
                "input_scaling": input_scaling,
                "fb_scaling": fb_scaling,
                "rc_connectivity": rc_connectivity,
                "input_connectivity": input_connectivity,
                "fb_connectivity": fb_connectivity,
                "noise_in": noise_in,
                "noise_rc": noise_rc,
                "noise_out": noise_fb,
                "noise_type": noise_type,
                "activation_type": activation,
                "activation": partial(
                    ip_activation, reservoir=self, f=get_function(activation)
                ),
                "fb_activation": fb_activation,
                "units": units,
                "noise_generator": partial(noise, rng=rng, **noise_kwargs),
            },
            forward=forward_external,
            initializer=partial(
                initialize,
                input_bias=input_bias,
                bias_scaling=bias_scaling,
                sr=sr,
                input_scaling=input_scaling,
                input_connectivity=input_connectivity,
                rc_connectivity=rc_connectivity,
                W_init=W,
                Win_init=Win,
                bias_init=bias,
                seed=seed,
            ),
            backward=backward,
            output_dim=units,
            feedback_dim=feedback_dim,
            name=name,
            **kwargs,
        )

    # TODO: handle unsupervised learners with a specific attribute
    @property
    def fitted(self):
        return True

    def partial_fit(self, X_batch, Y_batch=None, warmup=0, **kwargs) -> "Node":
        """Partial offline fitting method of a Node.
        Can be used to perform batched fitting or to pre-compute some variables
        used by the fitting method.

        Parameters
        ----------
        X_batch : array-like of shape ([series], timesteps, features)
            A sequence or a batch of sequence of input data.
        Y_batch : array-like of shape ([series], timesteps, features), optional
            A sequence or a batch of sequence of teacher signals.
        warmup : int, default to 0
            Number of timesteps to consider as warmup and
            discard at the beginning of each timeseries before training.

        Returns
        -------
            Node
                Partially fitted Node.
        """
        X, _ = check_xy(self, X_batch, allow_n_inputs=False)

        X, _ = _init_with_sequences(self, X)

        self.initialize_buffers()

        for i in range(len(X)):
            X_seq = X[i]

            if X_seq.shape[0] <= warmup:
                raise ValueError(
                    f"Warmup set to {warmup} timesteps, but one timeseries is only "
                    f"{X_seq.shape[0]} long."
                )

            if warmup > 0:
                self.run(X_seq[:warmup])

            self._partial_backward(self, X_seq[warmup:])

        return self



import json
import os
import time

import dill
import numpy as np
from scipy import sparse

from ..._version import __version__
from .. import regression_models


def _save(esn, directory: str):
    """Base utility for saving an ESN model, based on the ESN class.

    Arguments:
        esn {ESN} -- ESN model to save.
        directory {str or Path} -- Directory to store the model.
    """
    # create new directory
    savedir = os.path.join(directory)
    if not os.path.isdir(savedir):
        os.mkdir(savedir)
    else:
        raise OSError(f"Directory '{savedir}' already exists.")

    current_time = time.time()

    # store matrices
    Win_path = f"esn-Win-{current_time}.npy"
    Wout_path = None
    Wfb_path = None

    if sparse.issparse(esn.W):
        W_path = f"esn-W-{current_time}.npz"
        sparse.save_npz(os.path.join(savedir, W_path), esn.W)
    else:
        W_path = f"esn-W-{current_time}.npy"
        np.save(os.path.join(savedir, W_path), esn.W)
    np.save(os.path.join(savedir, Win_path), esn.Win)

    dim_out = None
    if esn.Wout is not None:
        Wout_path = f"esn-Wout-{current_time}.npy"
        np.save(os.path.join(savedir, Wout_path), esn.Wout)
        dim_out = esn.dim_out

    fbfunc = None
    fbfunc_info = None
    if esn.Wfb is not None:
        Wfb_path = f"esn-Wfb-{current_time}.npy"
        np.save(os.path.join(savedir, Wfb_path), esn.Wfb)
        # fbfunc is serialized and stored
        fbfunc_info = {"cls": str(esn.fbfunc.__class__), "name": esn.fbfunc.__name__}
        fbfunc = f"fbfunc_save-{current_time}"
        dim_out = esn.dim_out
        with open(os.path.join(savedir, fbfunc), "wb+") as f:
            dill.dump(esn.fbfunc, f)

    sklearn_model = None
    if getattr(esn.model, "model", None) is not None:
        sklearn_model = f"sklearn_func_save-{current_time}"
        # scikit-learn model is serialized and stored
        # will require scikit-learn to be imported when loading ESN.
        with open(os.path.join(savedir, sklearn_model), "wb+") as f:
            dill.dump(esn.model.model, f)

    # a copy of the ESN class is also serialized.
    # allow to load an ESN without necessarily using
    # the same version of Reservoirpy.
    cls_path = f"cls_bin-{current_time}"
    with open(os.path.join(savedir, cls_path), "wb+") as f:
        dill.dump(esn.__class__, f)

    attr = {
        "cls": esn.__class__.__name__,
        "cls_bin": cls_path,
        "version": __version__,
        "serial": current_time,
        "attr": {
            "_W": W_path,
            "_Win": Win_path,
            "_Wfb": Wfb_path,
            "_Wout": Wout_path,
            "_N": esn._N,
            "lr": esn.lr,
            "_input_bias": esn.input_bias,
            "_dim_in": esn._dim_in,
            "_dim_out": dim_out,
            "_ridge": esn.ridge,
            "typefloat": esn.typefloat.__name__,
            "sklearn_model": sklearn_model,
            "fbfunc": fbfunc,
            "noise_in": esn.noise_in,
            "noise_out": esn.noise_out,
            "noise_rc": esn.noise_rc,
            "seed": esn.seed,
            "model": esn.model.__class__.__name__,
        },
        "misc": {"fbfunc_info": fbfunc_info},
    }

    # save a summary file
    with open(os.path.join(savedir, "esn.json"), "w+") as f:
        json.dump(attr, f)


def _new_from_save(base_cls, restored_attr):

    obj = object.__new__(base_cls)
    for name, attr in restored_attr.items():
        try:
            obj.__setattr__(name, attr)
        except AttributeError as e:
            print(e)
            print(name, attr)

    obj.model = getattr(regression_models, obj.model)(obj._ridge, obj.sklearn_model)

    del obj.sklearn_model
    del obj._ridge

    return obj


def load(directory: str):
    """Load an ESN model in v0.2 format.

    Warning
    -------

        v0.2 models are deprecated. Consider using :py:func:`load_compat` to
        translate saved models from v0.2 to new Node API (see :ref:`node`)
        introduced in v0.3.

    Parameters
    ----------
        directory : str or Path
            Saved model directory.

    Returns
    -------
        :py:class:`compat.ESN`
            Loaded ESN.
    """
    with open(os.path.join(directory, "esn.json"), "r") as f:
        attr = json.load(f)

    model_attr = attr["attr"]

    if os.path.splitext(model_attr["_W"])[1] == ".npy":
        model_attr["_W"] = np.load(os.path.join(directory, model_attr["_W"]))
    else:
        model_attr["_W"] = sparse.load_npz(os.path.join(directory, model_attr["_W"]))

    model_attr["_Win"] = np.load(os.path.join(directory, model_attr["_Win"]))

    if model_attr["_Wout"] is not None:
        model_attr["_Wout"] = np.load(os.path.join(directory, model_attr["_Wout"]))
    if model_attr["_Wfb"] is not None:
        model_attr["_Wfb"] = np.load(os.path.join(directory, model_attr["_Wfb"]))
        with open(os.path.join(directory, model_attr["fbfunc"]), "rb") as f:
            model_attr["fbfunc"] = dill.load(f)

    elif model_attr["sklearn_model"] is not None:
        with open(os.path.join(directory, model_attr["sklearn_model"]), "rb") as f:
            model_attr["sklearn_model"] = dill.load(f)

    model_attr["typefloat"] = getattr(np, model_attr["typefloat"])

    with open(os.path.join(directory, attr["cls_bin"]), "rb") as f:
        base_cls = dill.load(f)

    if model_attr.get("activation") is None:
        model_attr["activation"] = np.tanh

    model = _new_from_save(base_cls, model_attr)

    return model



# Author: Deepayan Das at 16/08/2023 <deepayan.das@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from copy import deepcopy
from functools import partial

import numpy as np

from ...node import Node
from ...utils.random import rand_generator


def forward(readout: Node, X):
    instances = readout.params.get("instances")
    if type(instances) is not list:
        return instances.predict(X)
    else:
        return np.concatenate([instance.predict(X) for instance in instances], axis=-1)


def backward(readout: Node, X, Y):
    # Concatenate all the batches as one np.ndarray
    # of shape (timeseries*timesteps, features)
    X_ = np.concatenate(X, axis=0)
    Y_ = np.concatenate(Y, axis=0)

    instances = readout.params.get("instances")
    if type(instances) is not list:
        if readout.output_dim > 1:
            # Multi-output node and multi-output sklearn model
            instances.fit(X_, Y_)
        else:
            # Y_ should have 1 feature so we reshape to
            # (timeseries, ) to avoid scikit-learn's DataConversionWarning
            instances.fit(X_, Y_[..., 0])
    else:
        for i, instance in enumerate(instances):
            instance.fit(X_, Y_[..., i])


def initialize(readout: Node, x=None, y=None, model_hypers=None):
    if x is not None:
        in_dim = x.shape[1]
        if readout.output_dim is not None:
            out_dim = readout.output_dim
        elif y is not None:
            out_dim = y.shape[1]
        else:
            raise RuntimeError(
                f"Impossible to initialize {readout.name}: "
                f"output dimension was not specified at "
                f"creation, and no teacher vector was given."
            )

        readout.set_input_dim(in_dim)
        readout.set_output_dim(out_dim)

        first_instance = readout.model(**deepcopy(model_hypers))
        # If there are multiple output but the specified model doesn't support
        # multiple outputs, we create an instance of the model for each output.
        if out_dim > 1 and not first_instance._get_tags().get("multioutput"):
            instances = [
                readout.model(**deepcopy(model_hypers)) for i in range(out_dim)
            ]
            readout.set_param("instances", instances)
        else:
            readout.set_param("instances", first_instance)

        return


class ScikitLearnNode(Node):
    """
    A node interfacing a scikit-learn linear model that can be used as an offline
    readout node.

    The ScikitLearnNode takes a scikit-learn model as parameter and creates a
    node with the specified model.

    We currently support classifiers (like
    :py:class:`sklearn.linear_model.LogisticRegression` or
    :py:class:`sklearn.linear_model.RidgeClassifier`) and regressors (like
    :py:class:`sklearn.linear_model.Lasso` or
    :py:class:`sklearn.linear_model.ElasticNet`).

    For more information on the above-mentioned estimators,
    please visit scikit-learn linear model API reference
    <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.linear_model>`_

    :py:attr:`ScikitLearnNode.params` **list**

    ================== =================================================================
    ``instances``      Instance(s) of the model class used to fit and predict. If
                       :py:attr:`ScikitLearnNode.output_dim` > 1 and the model doesn't
                       support multi-outputs, `instances` is a list of instances, one
                       for each output feature.
    ================== =================================================================

    :py:attr:`ScikitLearnNode.hypers` **list**

    ================== =================================================================
    ``model``          (class) Underlying scikit-learn model.
    ``model_hypers``   (dict) Keyword arguments for the scikit-learn model.
    ================== =================================================================

    Parameters
    ----------
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    model : str, optional
        Node name.
    model_hypers
        (dict) Additional keyword arguments for the scikit-learn model.

    Example
    -------
    >>> from reservoirpy.nodes import Reservoir, ScikitLearnNode
    >>> from sklearn.linear_model import Lasso
    >>> reservoir = Reservoir(units=100)
    >>> readout = ScikitLearnNode(model=Lasso, model_hypers={"alpha":1e-5})
    >>> model = reservoir >> readout
    """

    def __init__(self, model, model_hypers=None, output_dim=None, **kwargs):
        if model_hypers is None:
            model_hypers = {}

        if not hasattr(model, "fit"):
            model_name = model.__name__
            raise AttributeError(
                f"Specified model {model_name} has no method called 'fit'."
            )
        if not hasattr(model, "predict"):
            model_name = model.__name__
            raise AttributeError(
                f"Specified model {model_name} has no method called 'predict'."
            )

        # Ensure reproducibility
        # scikit-learn currently only supports RandomState
        if (
            model_hypers.get("random_state") is None
            and "random_state" in model.__init__.__kwdefaults__
        ):

            generator = rand_generator()
            model_hypers.update(
                {
                    "random_state": np.random.RandomState(
                        seed=generator.integers(1 << 32)
                    )
                }
            )

        super(ScikitLearnNode, self).__init__(
            hypers={"model": model, "model_hypers": model_hypers},
            params={"instances": None},
            forward=forward,
            backward=backward,
            output_dim=output_dim,
            initializer=partial(initialize, model_hypers=model_hypers),
            **kwargs,
        )



# Author: Nathan Trouvain at 24/02/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import os
import uuid
from functools import partial
from multiprocessing import Manager, Process
from typing import Tuple

import joblib
import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm

from ...utils.parallel import clean_tempfile, get_joblib_backend, temp_registry


def parallelize(
    esn,
    func,
    workers,
    lengths,
    return_states,
    pbar_text=None,
    verbose=False,
    **func_kwargs,
):
    workers = min(len(lengths), workers)
    backend = get_joblib_backend() if workers > 1 or workers == -1 else "sequential"

    steps = np.sum(lengths)
    ends = np.cumsum(lengths)
    starts = ends - np.asarray(lengths)

    fn_kwargs = (
        {k: func_kwargs[k][i] for k in func_kwargs.keys()} for i in range(len(lengths))
    )

    states = None
    if return_states:
        shape = (steps, esn.N)
        states = memmap(shape, dtype=esn.typefloat, caller=esn)

    with ParallelProgressQueue(total=steps, text=pbar_text, verbose=verbose) as pbar:

        func = partial(func, pbar=pbar)

        with Parallel(backend=backend, n_jobs=workers) as parallel:

            def func_wrapper(states, start_pos, end_pos, *args, **kwargs):
                s = func(*args, **kwargs)

                out = None
                # if function returns states and outputs
                if hasattr(s, "__len__") and len(s) == 2:
                    out = s[0]  # outputs are always returned first
                    s = s[1]

                if return_states:
                    states[start_pos:end_pos] = s[:]

                return out

            outputs = parallel(
                delayed(func_wrapper)(states, start, end, **kwargs)
                for start, end, kwargs in zip(starts, ends, fn_kwargs)
            )

    if return_states:
        states = [np.array(states[start:end]) for start, end in zip(starts, ends)]

    clean_tempfile(esn)

    return outputs, states


class ParallelProgressQueue:
    def __init__(self, total, text, verbose):
        self._verbose = verbose
        if verbose is True:
            self._queue = Manager().Queue()
            self._process = Process(target=self._listen)
        else:
            self._queue = None
            self._process = None

        self._total = total
        self._text = text

    def __enter__(self):
        if self._verbose:
            self._process.start()
        return _ProgressBarQueue(self._queue, self._verbose)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._verbose:
            self._queue.put(None)
            self._process.join()

    def _listen(self):
        with tqdm(total=self._total, desc=self._text) as pbar:
            for i in iter(self._queue.get, None):
                pbar.update(i)


class _ProgressBarQueue:
    def __init__(self, queue, verbose):
        self._queue = queue
        self._verbose = verbose

    def update(self, value):
        if self._verbose:
            self._queue.put(value)


def memmap(shape: Tuple, dtype: np.dtype, mode: str = "w+", caller=None) -> np.memmap:
    """Create a new numpy.memmap object, stored in a temporary
    folder on disk.

    Parameters
    ----------
    shape : tuple of int
        Shape of the memmaped array.
    directory: str:
        Directory where the memmap will be stored.
    dtype : numpy.dtype
        Data type of the array.
    mode : {‘r+’, ‘r’, ‘w+’, ‘c’}, optional
        Mode in which to open the memmap file. See `Numpy documentation
        <https://numpy.org/doc/stable/reference/generated/numpy.memmap.html>`_
        for more information.

    Returns
    -------
        numpy.memmap
            An empty memory-mapped array.

    """
    global temp_registry
    from ... import _TEMPDIR

    if caller is not None:
        caller_name = caller.__class__.__name__
    else:
        caller_name = ""
    filename = os.path.join(_TEMPDIR, f"{caller_name + str(uuid.uuid4())}.dat")
    if caller is not None:
        temp_registry[caller].append(filename)
    return np.memmap(filename, shape=shape, mode=mode, dtype=dtype)


def as_memmap(data, caller=None):
    global temp_registry
    from ... import _TEMPDIR

    if caller is not None:
        caller_name = caller.__class__.__name__
    else:
        caller_name = ""
    filename = os.path.join(_TEMPDIR, f"{caller_name + str(uuid.uuid4())}.dat")
    joblib.dump(data, filename)
    temp_registry[caller].append(filename)
    return joblib.load(filename, mmap_mode="r+")
